@@ -35,7 +35,6 @@
  */
 
 #include <arpa/inet.h>
-#include <ldns/ldns.h>
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -46,29 +45,28 @@
 #include <netdb.h>
 
 #include "config.h"
+#include "gldns/str2wire.h"
 #include "context.h"
 #include "types-internal.h"
 #include "util-internal.h"
 #include "dnssec.h"
 #include "stub.h"
+#include "list.h"
 
 void *plain_mem_funcs_user_arg = MF_PLAIN;
 
 typedef struct host_name_addrs {
 	getdns_rbnode_t node;
-	ldns_rdf *host_name;
-	ldns_rr_list *ipv4addrs;
-	ldns_rr_list *ipv6addrs;
+	getdns_list *ipv4addrs;
+	getdns_list *ipv6addrs;
+	uint8_t host_name[];
 } host_name_addrs;
 
 /* Private functions */
 getdns_return_t create_default_namespaces(struct getdns_context *context);
-static void create_local_hosts(struct getdns_context *context);
-getdns_return_t destroy_local_hosts(struct getdns_context *context);
 static struct getdns_list *create_default_root_servers(void);
 static getdns_return_t set_os_defaults(struct getdns_context *);
 static int transaction_id_cmp(const void *, const void *);
-static int local_host_cmp(const void *, const void *);
 static void dispatch_updated(struct getdns_context *, uint16_t);
 static void cancel_dns_req(getdns_dns_req *);
 static void cancel_outstanding_requests(struct getdns_context*, int);
@@ -91,9 +89,8 @@ static void destroy_local_host(getdns_rbnode_t * node, void *arg)
 {
 	getdns_context *context = (getdns_context *)arg;
 	host_name_addrs *hnas = (host_name_addrs *)node;
-	ldns_rdf_deep_free(hnas->host_name);
-	ldns_rr_list_deep_free(hnas->ipv4addrs);
-	ldns_rr_list_deep_free(hnas->ipv6addrs);
+	getdns_list_destroy(hnas->ipv4addrs);
+	getdns_list_destroy(hnas->ipv6addrs);
 	GETDNS_FREE(context->my_mf, hnas);
 }
 
@@ -115,62 +112,277 @@ create_default_namespaces(struct getdns_context *context)
 	return GETDNS_RETURN_GOOD;
 }
 
-/**
- * Helper to get contents from hosts file
- */
-static void
-create_local_hosts(struct getdns_context *context)
+static inline void canonicalize_dname(uint8_t *dname)
 {
-	ldns_rr_list *host_names = ldns_get_rr_list_hosts_frm_file(NULL);
-	size_t i;
-	ldns_rr *rr;
-	host_name_addrs *hnas;
+	uint8_t *next_label;
 
-	if (host_names == NULL)
+	while (*dname) {
+		next_label = dname + *dname + 1;
+		dname += 1;
+		while (dname < next_label) {
+			*dname = (uint8_t)tolower((unsigned char)*dname);
+			dname++;
+		}
+	}
+}
+
+static int
+canonical_dname_compare(register const uint8_t *d1, register const uint8_t *d2)
+{
+	register uint8_t lab1, lab2;
+
+	assert(d1 && d2);
+
+	lab1 = *d1++;
+	lab2 = *d2++;
+	while (lab1 != 0 || lab2 != 0) {
+		/* compare label length */
+		/* if one dname ends, it has labellength 0 */
+		if (lab1 != lab2) {
+			if (lab1 < lab2)
+				return -1;
+			return 1;
+		}
+		while (lab1--) {
+			/* compare bytes first for speed */
+			if (*d1 != *d2) {
+				if (*d1 < *d2)
+					return -1;
+				return  1;
+			}
+			d1++;
+			d2++;
+		}
+		/* next pair of labels. */
+		lab1 = *d1++;
+		lab2 = *d2++;
+	}
+	return 0;
+}
+
+static int
+local_host_cmp(const void *id1, const void *id2)
+{
+	return canonical_dname_compare(id1, id2);
+}
+
+static void
+add_local_host(getdns_context *context, getdns_dict *address, const char *str)
+{
+	uint8_t host_name[256];
+	size_t host_name_len = sizeof(host_name);
+	host_name_addrs *hnas;
+	getdns_bindata *address_type;
+	int hnas_found = 0;
+	getdns_list **addrs;
+
+	if (gldns_str2wire_dname_buf(str, host_name, &host_name_len))
 		return;
 
-	/* We have a 1:1 list of name -> ip address where there is an 
-	   underlying many to many relationship. Need to create a lookup of
-	   (unique name + A/AAAA)-> list of IPV4/IPv6 ip addresses*/
-	for (i = 0; i < ldns_rr_list_rr_count(host_names); i++) {
-		rr = ldns_rr_list_rr(host_names, i);
+	canonicalize_dname(host_name);
+	
+	if (!(hnas = (host_name_addrs *)getdns_rbtree_search(
+	    &context->local_hosts, host_name))) {
 
-		if (ldns_rr_get_type(rr) != LDNS_RR_TYPE_A &&
-		    ldns_rr_get_type(rr) != LDNS_RR_TYPE_AAAA)
-			continue;
+		if (!(hnas = (host_name_addrs *)GETDNS_XMALLOC(context->mf,
+		    uint8_t, sizeof(host_name_addrs) + host_name_len)))
+			return;
 
-		/*Check to see if we already have an entry*/
-		hnas = (host_name_addrs *)getdns_rbtree_search(
-		   &context->local_hosts, ldns_rr_owner(rr));
-		if (!hnas) {
-			if (!(hnas = GETDNS_MALLOC(
-			    context->my_mf, host_name_addrs)))
-				break;
+		hnas->ipv4addrs = NULL;
+		hnas->ipv6addrs = NULL;
+		(void)memcpy(hnas->host_name, host_name, host_name_len);
+		hnas->node.key = &hnas->host_name;
 
-			if (!(hnas->host_name =
-			    ldns_rdf_clone(ldns_rr_owner(rr))) ||
-			    !(hnas->ipv4addrs = ldns_rr_list_new()) ||
-			    !(hnas->ipv6addrs = ldns_rr_list_new())) {
+	} else
+		hnas_found = 1;
+	
+	if (getdns_dict_get_bindata(address, "address_type", &address_type) ||
 
-				ldns_rdf_free(hnas->host_name);
-				ldns_rr_list_free(hnas->ipv4addrs);
-				ldns_rr_list_free(hnas->ipv6addrs);
-				GETDNS_FREE(context->my_mf, hnas);
-				break;
+	    address_type->size < 4 ||
+
+	    !(addrs = address_type->data[3] == '4'? &hnas->ipv4addrs
+	            : address_type->data[3] == '6'? &hnas->ipv4addrs : NULL)) {
+
+		if (!hnas_found) GETDNS_FREE(context->mf, hnas);
+		return;
+	}
+	if (!*addrs && !(*addrs = getdns_list_create_with_context(context))) {
+		if (!hnas_found) GETDNS_FREE(context->mf, hnas);
+		return;
+	}
+	if (getdns_list_append_dict(*addrs, address) && !hnas_found) {
+		getdns_list_destroy(*addrs);
+		GETDNS_FREE(context->mf, hnas);
+
+	} else if (!hnas_found)
+		(void)getdns_rbtree_insert(&context->local_hosts, &hnas->node);
+}
+
+static getdns_dict *
+sockaddr_dict(getdns_context *context, struct sockaddr *sa)
+{
+	getdns_dict *address = getdns_dict_create_with_context(context);
+	char addrstr[1024], *b;
+	getdns_bindata bindata;
+	uint16_t port;
+
+	if (!address)
+		return NULL;
+
+	switch (sa->sa_family) {
+	case AF_INET:
+		if (getdns_dict_util_set_string(address,"address_type","IPv4"))
+			break;
+
+		bindata.size = 4;
+		bindata.data = (void *)&((struct sockaddr_in*)sa)->sin_addr;
+		if ((getdns_dict_set_bindata(address,"address_data",&bindata)))
+			break;
+
+		port = ntohs(((struct sockaddr_in *)sa)->sin_port);
+		if (port !=  0 && port != 53 &&
+		    getdns_dict_set_int(address, "port", (uint32_t)port))
+			break;
+
+		return address;
+
+	case AF_INET6:
+		if (getdns_dict_util_set_string(address,"address_type","IPv6"))
+			break;
+
+		bindata.size = 16;
+		bindata.data = (void *)&((struct sockaddr_in6*)sa)->sin6_addr;
+		if ((getdns_dict_set_bindata(address,"address_data",&bindata)))
+			break;
+
+		port = ntohs(((struct sockaddr_in6 *)sa)->sin6_port);
+		if (port !=  0 && port != 53 &&
+		    getdns_dict_set_int(address, "port", (uint32_t)port))
+			break;
+
+		/* Try to get scope_id too */
+		if (getnameinfo(sa, sizeof(struct sockaddr_in6),
+		    addrstr, sizeof(addrstr), NULL, 0, NI_NUMERICHOST))
+			break;
+		if ((b = strchr(addrstr, '%')) &&
+		    getdns_dict_util_set_string(address, "scope_id", b+1))
+			break;
+
+		return address;
+	default:
+		/* Unknown protocol */
+		break;
+	}
+	getdns_dict_destroy(address);
+	return NULL;
+}
+
+static getdns_dict *
+str_addr_dict(getdns_context *context, const char *str)
+{
+	static struct addrinfo hints = { .ai_family = AF_UNSPEC
+	                               , .ai_flags  = AI_NUMERICHOST };
+	struct addrinfo *ai;
+	getdns_dict *address;
+
+	if (getaddrinfo(str, NULL, &hints, &ai))
+		return NULL;
+
+	address = sockaddr_dict(context, ai->ai_addr);
+	freeaddrinfo(ai);
+
+	return address;
+}
+
+static void
+create_local_hosts(getdns_context *context)
+{
+	/* enough space in buf for longest allowed domain name */
+	char buf[1024];
+	char *pos = buf, prev_c, *start_of_word = NULL;
+	FILE *in;
+	int start_of_line = 1;
+	getdns_dict *address = NULL;
+
+	in = fopen("/etc/hosts", "r");
+	while (fgets(pos, (int)(sizeof(buf) - (pos - buf)), in)) {
+		pos = buf;
+		/* Break out of for to read more */
+		for (;;) {
+			/* Skip whitespace */
+			while (*pos == ' '  || *pos == '\f' 
+			    || *pos == '\t' || *pos == '\v')
+				pos++;
+
+			if (*pos == '\0') { /* End of read data */
+				pos = buf;
+				goto read_more;
+
+			} else if (*pos == '#' || *pos == '\r' || *pos == '\n')
+				/* Comments or end of line */
+				break; /* skip to next line */
+
+			assert(*pos && !isspace(*pos));
+
+			start_of_word = pos;
+
+			/* Search for end of word */
+			while (*pos && !isspace(*pos))
+				pos++;
+
+			/* '\0' before whitespace, so either the word did not
+			 * fit, or we are at the end of the file.
+			 */
+			if (*pos == '\0') {
+				if (start_of_word == buf) /* word too big */
+					break; /* skip to next line */
+
+				/* Move word to fit in buffer */
+				memmove(buf,start_of_word,pos - start_of_word);
+				pos = buf + (pos - start_of_word);
+				start_of_word = buf;
+				*pos = '\0';
+				goto read_more;
 			}
-			hnas->node.key = hnas->host_name;
-			(void) getdns_rbtree_insert(
-			    &context->local_hosts, &hnas->node);
+			assert(isspace(*pos));
+			prev_c = *pos;
+			*pos = '\0';
+			if (start_of_line) {
+				start_of_line = 0;
+				if (address) 
+					getdns_dict_destroy(address);
+				if (!(address =
+				    str_addr_dict(context, start_of_word)))
+					/* Unparseable address */
+					break; /* skip to next line */
+			} else 
+				add_local_host(context, address, start_of_word);
+
+			start_of_word = NULL;
+			*pos = prev_c;
+			/* process next word in buf */
 		}
-		if (!(rr = ldns_rr_clone(rr)))
-			break;
-		if (ldns_rr_get_type(rr) == LDNS_RR_TYPE_A) {
-			if (!ldns_rr_list_push_rr(hnas->ipv4addrs, rr))
-				ldns_rr_free(rr);
-		} else if (!ldns_rr_list_push_rr(hnas->ipv6addrs, rr))
-			ldns_rr_free(rr);
+		/* skip to next line */
+		while (*pos != '\r' && *pos != '\n')
+			if (*pos)
+				pos++;
+			else if (!fgets((pos = buf), sizeof(buf), in))
+				break; /* We're done */
+		start_of_line = 1;
+		if (address) {
+			getdns_dict_destroy(address);
+			address = NULL;
+		}
+		pos = buf;
+read_more:	;
+	}
+	fclose(in);
+	if (address) {
+		/* One last name for this address? */
+		if (start_of_word && !start_of_line)
+			add_local_host(context, address, start_of_word);
+		getdns_dict_destroy(address);
 	}
-	ldns_rr_list_deep_free(host_names);
 }
 
 /**
@@ -258,17 +470,22 @@ upstreams_resize(getdns_upstreams *upstreams, size_t size)
 	return r;
 }
 
-static void
-upstreams_dereference(getdns_upstreams *upstreams)
+void
+priv_getdns_upstreams_dereference(getdns_upstreams *upstreams)
 {
-	if (upstreams && --upstreams->referenced == 0)
-		GETDNS_FREE(upstreams->mf, upstreams);
-}
+	size_t i;
 
-static size_t
-upstream_addr_len(getdns_upstream *upstream)
-{
-	return upstream->addr.ss_family == AF_INET ? 4 : 16;
+	if (upstreams && --upstreams->referenced == 0) {
+		for (i = 0; i < upstreams->count; i++) {
+			if (upstreams->upstreams[i].tls_obj != NULL) {
+				SSL_shutdown(upstreams->upstreams[i].tls_obj);
+				SSL_free(upstreams->upstreams[i].tls_obj);
+			}
+			if (upstreams->upstreams[i].fd != -1)
+				close(upstreams->upstreams[i].fd);
+		}
+		GETDNS_FREE(upstreams->mf, upstreams);
+	}
 }
 
 static uint8_t*
@@ -297,7 +514,8 @@ upstream_scope_id(getdns_upstream *upstream)
 }
 
 static void
-upstream_ntop_buf(getdns_upstream *upstream, char *buf, size_t len)
+upstream_ntop_buf(getdns_upstream *upstream, getdns_transport_t transport,
+                  char *buf, size_t len)
 {
 	/* Also possible but prints scope_id by name (nor parsed by unbound)
 	 *
@@ -309,36 +527,14 @@ upstream_ntop_buf(getdns_upstream *upstream, char *buf, size_t len)
 	if (upstream_scope_id(upstream))
 		(void) snprintf(buf + strlen(buf), len - strlen(buf),
 		    "%%%d", (int)*upstream_scope_id(upstream));
-	if (upstream_port(upstream) != 53 && upstream_port(upstream) != 0)
+	if (transport == GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN)
+		(void) snprintf(buf + strlen(buf), len - strlen(buf),
+		    "@%d", GETDNS_TLS_PORT);
+	else if (upstream_port(upstream) != 53 && upstream_port(upstream) != 0)
 		(void) snprintf(buf + strlen(buf), len - strlen(buf),
 		    "@%d", (int)upstream_port(upstream));
 }
 
-static getdns_dict *
-upstream_dict(getdns_context *context, getdns_upstream *upstream)
-{
-	getdns_dict *r = getdns_dict_create_with_context(context);
-	char addrstr[1024], *b;
-	getdns_bindata bindata;
-
-	getdns_dict_util_set_string(r, "address_type",
-	    upstream->addr.ss_family == AF_INET ? "IPv4" : "IPv6");
-
-	bindata.size = upstream_addr_len(upstream);
-	bindata.data = upstream_addr(upstream);
-	getdns_dict_set_bindata(r, "address_data", &bindata);
-
-	if (upstream_port(upstream) != 53)
-		getdns_dict_set_int(r, "port", upstream_port(upstream));
-
-	(void) getnameinfo((struct sockaddr *)&upstream->addr,
-	    upstream->addr_len, addrstr, 1024, NULL, 0, NI_NUMERICHOST);
-	if ((b = strchr(addrstr, '%')))
-		getdns_dict_util_set_string(r, "scope_id", b+1);
-
-	return r;
-}
-
 static int
 net_req_query_id_cmp(const void *id1, const void *id2)
 {
@@ -360,6 +556,7 @@ upstream_init(getdns_upstream *upstream,
 
 	/* For sharing a socket to this upstream with TCP  */
 	upstream->fd       = -1;
+	upstream->tls_obj  = NULL;
 	upstream->loop = NULL;
 	(void) getdns_eventloop_event_init(
 	    &upstream->event, upstream, NULL, NULL, NULL);
@@ -368,15 +565,15 @@ upstream_init(getdns_upstream *upstream,
 	upstream->write_queue = NULL;
 	upstream->write_queue_last = NULL;
 
+	upstream->has_client_cookie = 0;
+	upstream->has_prev_client_cookie = 0;
+	upstream->has_server_cookie = 0;
+
 	/* Tracking of network requests on this socket */
 	getdns_rbtree_init(&upstream->netreq_by_query_id,
 	    net_req_query_id_cmp);
 }
 
-/*---------------------------------------- set_os_defaults
-  we use ldns to read the resolv.conf file - the ldns resolver is
-  destroyed once the file is read
-*/
 static getdns_return_t
 set_os_defaults(struct getdns_context *context)
 {
@@ -408,7 +605,7 @@ set_os_defaults(struct getdns_context *context)
 	in = fopen(context->fchg_resolvconf->fn, "r");
 	if (!in)
 		return GETDNS_RETURN_GOOD;
-	
+
 	memset(&hints, 0, sizeof(struct addrinfo));
 	hints.ai_family    = AF_UNSPEC;      /* Allow IPv4 or IPv6 */
 	hints.ai_socktype  = 0;              /* Datagram socket */
@@ -516,12 +713,10 @@ transaction_id_cmp(const void *id1, const void *id2)
     }
 }
 
-static int
-local_host_cmp(const void *id1, const void *id2)
-{
-	return (ldns_rdf_compare((const ldns_rdf *)id1,
-                                 (const ldns_rdf *)id2));
-}
+static void
+NULL_update_callback(
+    getdns_context *context, getdns_context_code_t code, void *userarg)
+{ (void)context; (void)code; (void)userarg; }
 
 /*
  * getdns_context_create
@@ -561,7 +756,9 @@ getdns_context_create_with_extended_memory_functions(
 	result->my_mf.mf.ext.realloc = realloc;
 	result->my_mf.mf.ext.free    = free;
 
-	result->update_callback = NULL;
+	result->update_callback  = NULL;
+	result->update_callback2 = NULL_update_callback;
+	result->update_userarg   = NULL;
 
 	result->mf.mf_arg         = userarg;
 	result->mf.mf.ext.malloc  = malloc;
@@ -589,6 +786,7 @@ getdns_context_create_with_extended_memory_functions(
 	result->edns_extended_rcode = 0;
 	result->edns_version = 0;
 	result->edns_do_bit = 0;
+	result-> tls_ctx = NULL;
 
 	result->extension = &result->mini_event.loop;
 	if ((r = getdns_mini_event_init(result, &result->mini_event)))
@@ -608,6 +806,9 @@ getdns_context_create_with_extended_memory_functions(
 	result->return_dnssec_status = GETDNS_EXTENSION_FALSE;
 
 	/* unbound context is initialized here */
+	/* Unbound needs SSL to be init'ed this early when TLS is used. However we
+	 * don't know that till later so we will have to do this every time. */
+	SSL_library_init();
 	result->unbound_ctx = NULL;
 	if ((r = rebuild_ub_ctx(result)))
 		goto error;
@@ -695,6 +896,9 @@ getdns_context_destroy(struct getdns_context *context)
 			GETDNS_FREE(context->my_mf, context->fchg_hosts->prevstat);
 		GETDNS_FREE(context->my_mf, context->fchg_hosts);
 	}
+	if (context->tls_ctx) {
+		SSL_CTX_free(context->tls_ctx);
+	}
 
     getdns_list_destroy(context->dns_root_servers);
     getdns_list_destroy(context->suffix);
@@ -706,7 +910,7 @@ getdns_context_destroy(struct getdns_context *context)
 
 	getdns_traverse_postorder(&context->local_hosts,
 	    destroy_local_host, context);
-	upstreams_dereference(context->upstreams);
+	priv_getdns_upstreams_dereference(context->upstreams);
 
     GETDNS_FREE(context->my_mf, context);
 }               /* getdns_context_destroy */
@@ -725,10 +929,31 @@ getdns_context_set_context_update_callback(struct getdns_context *context,
     return GETDNS_RETURN_GOOD;
 }               /* getdns_context_set_context_update_callback */
 
+getdns_return_t
+getdns_context_set_update_callback(getdns_context *context, void *userarg,
+    void (*cb)(getdns_context *, getdns_context_code_t, void *))
+{
+	if (!context) return GETDNS_RETURN_INVALID_PARAMETER;
+	context->update_userarg = userarg;
+	context->update_callback2 = cb ? cb : NULL_update_callback;
+	return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_update_callback(getdns_context *context, void **userarg,
+    void (**cb)(getdns_context *, getdns_context_code_t, void *))
+{
+	if (!context || !userarg || !cb)
+		return GETDNS_RETURN_INVALID_PARAMETER;
+
+	*userarg = context->update_userarg;
+	*cb = context->update_callback2;
+	return GETDNS_RETURN_GOOD;
+}
+
 /*
  * Helpers to set options on the unbound ctx
  */
-
 static void
 set_ub_string_opt(struct getdns_context *ctx, char *opt, char *value)
 {
@@ -829,6 +1054,10 @@ rebuild_ub_ctx(struct getdns_context* context) {
 static void
 dispatch_updated(struct getdns_context *context, uint16_t item)
 {
+	if (context->update_callback2 != NULL_update_callback)
+		context->update_callback2(
+		    context, item, context->update_userarg);
+
     if (context->update_callback) {
         context->update_callback(context, item);
     }
@@ -895,6 +1124,33 @@ getdns_context_set_namespaces(struct getdns_context *context,
     return GETDNS_RETURN_GOOD;
 }               /* getdns_context_set_namespaces */
 
+getdns_base_transport_t
+priv_get_base_transport(getdns_transport_t transport, int level) {
+	if (!(level == 0 || level == 1)) return GETDNS_TRANSPORT_NONE;
+	switch (transport) {
+ 		case GETDNS_TRANSPORT_UDP_FIRST_AND_FALL_BACK_TO_TCP:
+			if (level == 0) return GETDNS_TRANSPORT_UDP;
+			if (level == 1) return GETDNS_TRANSPORT_TCP;
+		case GETDNS_TRANSPORT_UDP_ONLY:
+			if (level == 0) return GETDNS_TRANSPORT_UDP;
+			if (level == 1) return GETDNS_TRANSPORT_NONE;
+		case GETDNS_TRANSPORT_TCP_ONLY:
+			if (level == 0) return GETDNS_TRANSPORT_TCP_SINGLE;
+			if (level == 1) return GETDNS_TRANSPORT_NONE;
+		case GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN:
+			if (level == 0) return GETDNS_TRANSPORT_TCP;
+			if (level == 1) return GETDNS_TRANSPORT_NONE;
+		case GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN:
+			if (level == 0) return GETDNS_TRANSPORT_TLS;
+			if (level == 1) return GETDNS_TRANSPORT_NONE;
+		case GETDNS_TRANSPORT_TLS_FIRST_AND_FALL_BACK_TO_TCP_KEEP_CONNECTIONS_OPEN:
+			if (level == 0) return GETDNS_TRANSPORT_TLS;
+			if (level == 1) return GETDNS_TRANSPORT_TCP;
+		default:
+			return GETDNS_TRANSPORT_NONE;
+		}
+}
+
 static getdns_return_t
 set_ub_dns_transport(struct getdns_context* context,
     getdns_transport_t value) {
@@ -908,13 +1164,25 @@ set_ub_dns_transport(struct getdns_context* context,
             set_ub_string_opt(context, "do-tcp:", "no");
             break;
         case GETDNS_TRANSPORT_TCP_ONLY:
-	case GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN:
+        /* Note: no pipelining available directly in unbound.*/
+        case GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN:
             set_ub_string_opt(context, "do-udp:", "no");
             set_ub_string_opt(context, "do-tcp:", "yes");
             break;
-        default:
-            /* TODO GETDNS_CONTEXT_TCP_ONLY_KEEP_CONNECTIONS_OPEN */
-            return GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
+       case GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN:
+            /* Hum. If used in recursive mode this will try TLS on port 53...
+             * So we need to fix or document that or delay setting it until 
+             * resolution.*/
+            set_ub_string_opt(context, "ssl-upstream:", "yes");
+            /* Fall through*/
+       case GETDNS_TRANSPORT_TLS_FIRST_AND_FALL_BACK_TO_TCP_KEEP_CONNECTIONS_OPEN:
+           /* Note: no fallback to TCP available directly in unbound, so we just
+            * use TCP for now to make sure the messages are sent. */
+           set_ub_string_opt(context, "do-udp:", "no");
+           set_ub_string_opt(context, "do-tcp:", "yes");
+           break;
+       default:
+           return GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
         }
     return GETDNS_RETURN_GOOD;
 }
@@ -928,6 +1196,11 @@ getdns_context_set_dns_transport(struct getdns_context *context,
     getdns_transport_t value)
 {
     RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    /* Note that the call below does not have any effect in unbound after the
+     * ctx is finalised. So will not apply for recursive mode or stub + dnssec.
+     * However the method returns success as otherwise the transport could not
+     * be reset for stub mode..... 
+     * Also, not all transport options supported in libunbound yet */
     if (set_ub_dns_transport(context, value) != GETDNS_RETURN_GOOD) {
         return GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
     }
@@ -1169,7 +1442,7 @@ getdns_context_set_upstream_recursive_servers(struct getdns_context *context,
 
 	RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
 	RETURN_IF_NULL(upstream_list, GETDNS_RETURN_INVALID_PARAMETER);
-	
+
 	r = getdns_list_get_length(upstream_list, &count);
 	if (count == 0 || r != GETDNS_RETURN_GOOD) {
 		return GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
@@ -1219,11 +1492,11 @@ getdns_context_set_upstream_recursive_servers(struct getdns_context *context,
 		if (inet_ntop(upstream->addr.ss_family, address_data->data,
 		    addrstr, 1024) == NULL)
 			goto invalid_parameter;
-		
+
 		port = 53;
 		(void) getdns_dict_get_int(dict, "port", &port);
 		(void) snprintf(portstr, 1024, "%d", (int)port);
-		
+
 		if (getdns_dict_get_bindata(dict, "scope_id", &scope_id) ==
 		    GETDNS_RETURN_GOOD) {
 			if (strlen(addrstr) + scope_id->size > 1022)
@@ -1241,7 +1514,7 @@ getdns_context_set_upstream_recursive_servers(struct getdns_context *context,
 		upstreams->count++;
 		freeaddrinfo(ai);
 	}
-	upstreams_dereference(context->upstreams);
+	priv_getdns_upstreams_dereference(context->upstreams);
 	context->upstreams = upstreams;
 	dispatch_updated(context,
 		GETDNS_CONTEXT_CODE_UPSTREAM_RECURSIVE_SERVERS);
@@ -1251,7 +1524,7 @@ getdns_context_set_upstream_recursive_servers(struct getdns_context *context,
 invalid_parameter:
 	r = GETDNS_RETURN_INVALID_PARAMETER;
 error:
-	upstreams_dereference(upstreams);
+	priv_getdns_upstreams_dereference(upstreams);
 	return GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
 } /* getdns_context_set_upstream_recursive_servers */
 
@@ -1389,14 +1662,14 @@ getdns_context_set_memory_functions(struct getdns_context *context,
 static void
 cancel_dns_req(getdns_dns_req *req)
 {
-	getdns_network_req *netreq;
-	
-	for (netreq = req->first_req; netreq; netreq = netreq->next)
+	getdns_network_req *netreq, **netreq_p;
+
+	for (netreq_p = req->netreqs; (netreq = *netreq_p); netreq_p++)
 		if (netreq->unbound_id != -1) {
 			ub_cancel(req->context->unbound_ctx,
 			    netreq->unbound_id);
 			netreq->unbound_id = -1;
-		} else 
+		} else
 			priv_getdns_cancel_stub_request(netreq);
 
 	req->canceled = 1;
@@ -1447,17 +1720,18 @@ getdns_cancel_callback(getdns_context *context,
 } /* getdns_cancel_callback */
 
 static getdns_return_t
-ub_setup_stub(struct ub_ctx *ctx, getdns_upstreams *upstreams)
+ub_setup_stub(struct ub_ctx *ctx, getdns_context *context)
 {
 	getdns_return_t r = GETDNS_RETURN_GOOD;
 	size_t i;
 	getdns_upstream *upstream;
 	char addr[1024];
+	getdns_upstreams *upstreams = context->upstreams;
 
 	(void) ub_ctx_set_fwd(ctx, NULL);
 	for (i = 0; i < upstreams->count; i++) {
 		upstream = &upstreams->upstreams[i];
-		upstream_ntop_buf(upstream, addr, 1024);
+		upstream_ntop_buf(upstream, context->dns_transport, addr, 1024);
 		ub_ctx_set_fwd(ctx, addr);
 	}
 
@@ -1526,16 +1800,10 @@ priv_getdns_ns_dns_setup(struct getdns_context *context)
 	assert(context);
 
 	switch (context->resolution_type) {
-	case GETDNS_RESOLUTION_STUB: 
-		/* Since we don't know if the resolution will be sync or async at this
-		 * point and we only support ldns in sync mode then we must set _both_
-		 * contexts up */
-		/* We get away with just setting up ldns here here because sync mode
-		 * always hits this method because at the moment all sync calls use DNS
-		 * namespace */
+	case GETDNS_RESOLUTION_STUB:
 		if (!context->upstreams || !context->upstreams->count)
 			return GETDNS_RETURN_GENERIC_ERROR;
-		return ub_setup_stub(context->unbound_ctx, context->upstreams);
+		return ub_setup_stub(context->unbound_ctx, context);
 
 	case GETDNS_RESOLUTION_RECURSING:
 		/* TODO: use the root servers via root hints file */
@@ -1556,6 +1824,33 @@ getdns_context_prepare_for_resolution(struct getdns_context *context,
     if (context->destroying) {
         return GETDNS_RETURN_BAD_CONTEXT;
     }
+
+	/* Transport can in theory be set per query in stub mode */
+	/* TODO: move this transport logic to a separate functions*/
+	if (context->resolution_type == GETDNS_RESOLUTION_STUB) {
+		switch (context->dns_transport) {
+		case GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN:
+		case GETDNS_TRANSPORT_TLS_FIRST_AND_FALL_BACK_TO_TCP_KEEP_CONNECTIONS_OPEN:
+			if (context->tls_ctx == NULL) {
+#ifdef HAVE_LIBTLS1_2
+				/* Create client context, use TLS v1.2 only for now */
+				context->tls_ctx = SSL_CTX_new(TLSv1_2_client_method());
+#endif
+				if(!context->tls_ctx && context->dns_transport == 
+				   GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN) {
+					return GETDNS_RETURN_BAD_CONTEXT;
+				}
+			}
+			break;
+		default:
+			break;
+		}
+	}
+	/* Block use of TLS ONLY in recursive mode as it won't work */
+	if (context->resolution_type == GETDNS_RESOLUTION_RECURSING
+	&& context->dns_transport == GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN)
+		return GETDNS_RETURN_BAD_CONTEXT;
+
 	if (context->resolution_type_set == context->resolution_type)
         	/* already set and no config changes
 		 * have caused this to be bad.
@@ -1840,7 +2135,8 @@ priv_get_context_settings(getdns_context* context) {
 		for (i = 0; i < context->upstreams->count; i++) {
 			getdns_dict *d;
 			upstream = &context->upstreams->upstreams[i];
-			d = upstream_dict(context, upstream);
+			d = sockaddr_dict(context,
+			    (struct sockaddr *)&upstream->addr);
 			r |= getdns_list_set_dict(upstreams, i, d);
 			getdns_dict_destroy(d);
 		}
@@ -1913,45 +2209,92 @@ getdns_context_set_use_threads(getdns_context* context, int use_threads) {
     return r == 0 ? GETDNS_RETURN_GOOD : GETDNS_RETURN_CONTEXT_UPDATE_FAIL;
 }
 
-getdns_return_t 
+getdns_return_t
 getdns_context_local_namespace_resolve(
     getdns_dns_req *dnsreq, getdns_dict **response)
 {
-	getdns_context *context = dnsreq->context;
-	ldns_rr_list *result_list = NULL;
+	getdns_context  *context = dnsreq->context;
 	host_name_addrs *hnas;
-	ldns_rdf *query_name;
-	int ipv4 = dnsreq->first_req->request_type == GETDNS_RRTYPE_A ||
-	    (dnsreq->first_req->next &&
-	     dnsreq->first_req->next->request_type == GETDNS_RRTYPE_A);
-	int ipv6 = dnsreq->first_req->request_type == GETDNS_RRTYPE_AAAA ||
-	    (dnsreq->first_req->next &&
-	     dnsreq->first_req->next->request_type == GETDNS_RRTYPE_AAAA);
-	
+	uint8_t query_name[256];
+	size_t  query_name_len = sizeof(query_name);
+	uint8_t lookup[256];
+	getdns_list    empty_list = { 0 };
+	getdns_bindata bindata;
+	getdns_list   *jaa;
+	size_t         i;
+	getdns_dict   *addr;
+	int ipv4 = dnsreq->netreqs[0]->request_type == GETDNS_RRTYPE_A ||
+	    (dnsreq->netreqs[1] &&
+	     dnsreq->netreqs[1]->request_type == GETDNS_RRTYPE_A);
+	int ipv6 = dnsreq->netreqs[0]->request_type == GETDNS_RRTYPE_AAAA ||
+	    (dnsreq->netreqs[1] &&
+	     dnsreq->netreqs[1]->request_type == GETDNS_RRTYPE_AAAA);
+
 	if (!ipv4 && !ipv6)
 		return GETDNS_RETURN_GENERIC_ERROR;
 
 	/*Do the lookup*/
-	if (!(query_name = ldns_rdf_new_frm_str(
-	    LDNS_RDF_TYPE_DNAME, dnsreq->name)))
-		return GETDNS_RETURN_MEMORY_ERROR;
-
-	if ((hnas = (host_name_addrs *)getdns_rbtree_search(
-	    &context->local_hosts, query_name))) {
-
-		result_list = ldns_rr_list_new();
-		if (ipv4)
-			(void) ldns_rr_list_cat(result_list, hnas->ipv4addrs);
-		if (ipv6)
-			(void) ldns_rr_list_cat(result_list, hnas->ipv6addrs);
-	}
-	ldns_rdf_deep_free(query_name);
-	if (!result_list)
+	if (gldns_str2wire_dname_buf(dnsreq->name,query_name,&query_name_len))
 		return GETDNS_RETURN_GENERIC_ERROR;
-	*response = create_getdns_response_from_rr_list(dnsreq, result_list);
-	/* Not deep_free because ldns_rr_list_cat doesn't clone the rr's */
-	ldns_rr_list_free(result_list);
-	return *response ? GETDNS_RETURN_GOOD : GETDNS_RETURN_GENERIC_ERROR;
+
+	(void)memcpy(lookup, query_name, query_name_len);
+	canonicalize_dname(lookup);
+
+	if (!(hnas = (host_name_addrs *)
+	    getdns_rbtree_search(&context->local_hosts, lookup)))
+		return GETDNS_RETURN_GENERIC_ERROR;
+
+	if (!hnas->ipv4addrs && (!ipv6 || !hnas->ipv6addrs))
+		return GETDNS_RETURN_GENERIC_ERROR;
+
+	if (!hnas->ipv6addrs && (!ipv4 || !hnas->ipv4addrs))
+		return GETDNS_RETURN_GENERIC_ERROR;
+
+	if (!(*response = getdns_dict_create_with_context(context)))
+		return GETDNS_RETURN_GENERIC_ERROR;
+
+	bindata.size = query_name_len;
+	bindata.data = query_name;
+	if (getdns_dict_set_bindata(*response, "canonical_name", &bindata))
+		goto error;
+
+	empty_list.mf = context->mf;
+	if (getdns_dict_set_list(*response, "replies_full", &empty_list))
+		goto error;
+
+	if (getdns_dict_set_list(*response, "replies_tree", &empty_list))
+		goto error;
+
+	if (getdns_dict_set_int(*response, "status", GETDNS_RESPSTATUS_GOOD))
+		goto error;
+
+	if (!ipv4 || !hnas->ipv4addrs) {
+		if (getdns_dict_set_list(*response,
+		    "just_address_answers", hnas->ipv6addrs))
+			goto error;
+		return GETDNS_RETURN_GOOD;
+	} else if (!ipv6 || !hnas->ipv6addrs) {
+		if (getdns_dict_set_list(*response,
+		    "just_address_answers", hnas->ipv4addrs))
+			goto error;
+		return GETDNS_RETURN_GOOD;
+	}
+	if (!(jaa = getdns_list_create_with_context(context)))
+		goto error;
+	for (i = 0; !getdns_list_get_dict(hnas->ipv4addrs, i, &addr); i++)
+		if (getdns_list_append_dict(jaa, addr))
+			break;
+	for (i = 0; !getdns_list_get_dict(hnas->ipv6addrs, i, &addr); i++)
+		if (getdns_list_append_dict(jaa, addr))
+			break;
+	if (!getdns_dict_set_list(*response, "just_address_answers", jaa)) {
+		getdns_list_destroy(jaa);
+		return GETDNS_RETURN_GOOD;
+	}
+	getdns_list_destroy(jaa);
+error:
+	getdns_dict_destroy(*response);
+	return GETDNS_RETURN_GENERIC_ERROR;
 }
 
 struct mem_funcs *
@@ -1960,4 +2303,182 @@ priv_getdns_context_mf(getdns_context *context)
 	return &context->mf;
 }
 
+/** begin getters **/
+getdns_return_t
+getdns_context_get_resolution_type(getdns_context *context,
+    getdns_resolution_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->resolution_type;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_namespaces(getdns_context *context,
+    size_t* namespace_count, getdns_namespace_t **namespaces) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(namespace_count, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(namespaces, GETDNS_RETURN_INVALID_PARAMETER);
+    *namespace_count = context->namespace_count;
+    if (!context->namespace_count) {
+        *namespaces = NULL;
+        return GETDNS_RETURN_GOOD;
+    }
+    // use normal malloc here so users can do normal free
+    *namespaces = malloc(context->namespace_count * sizeof(getdns_namespace_t));
+    memcpy(*namespaces, context->namespaces,
+           context->namespace_count * sizeof(getdns_namespace_t));
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_dns_transport(getdns_context *context,
+    getdns_transport_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->dns_transport;
+    return GETDNS_RETURN_GOOD;
+}
+
+
+getdns_return_t
+getdns_context_get_limit_outstanding_queries(getdns_context *context,
+    uint16_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->limit_outstanding_queries;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_timeout(getdns_context *context, uint64_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->timeout;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_follow_redirects(getdns_context *context,
+    getdns_redirects_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->follow_redirects;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_dns_root_servers(getdns_context *context,
+    getdns_list **value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = NULL;
+    if (context->dns_root_servers) {
+        return getdns_list_copy(context->dns_root_servers, value);
+    }
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_append_name(getdns_context *context,
+    getdns_append_name_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->append_name;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_suffix(getdns_context *context, getdns_list **value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = NULL;
+    if (context->suffix) {
+        return getdns_list_copy(context->suffix, value);
+    }
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_dnssec_trust_anchors(getdns_context *context,
+    getdns_list **value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = NULL;
+    if (context->dnssec_trust_anchors) {
+        return getdns_list_copy(context->dnssec_trust_anchors, value);
+    }
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_dnssec_allowed_skew(getdns_context *context,
+    uint32_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->dnssec_allowed_skew;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_upstream_recursive_servers(getdns_context *context,
+    getdns_list **upstream_list) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(upstream_list, GETDNS_RETURN_INVALID_PARAMETER);
+    *upstream_list = NULL;
+    if (context->upstreams && context->upstreams->count > 0) {
+        getdns_return_t r = GETDNS_RETURN_GOOD;
+        size_t i;
+        getdns_upstream *upstream;
+        getdns_list *upstreams = getdns_list_create();
+        for (i = 0; i < context->upstreams->count; i++) {
+            getdns_dict *d;
+            upstream = &context->upstreams->upstreams[i];
+            d = sockaddr_dict(context, (struct sockaddr *)&upstream->addr);
+            r |= getdns_list_set_dict(upstreams, i, d);
+            getdns_dict_destroy(d);
+        }
+        if (r != GETDNS_RETURN_GOOD) {
+            getdns_list_destroy(upstreams);
+            return GETDNS_RETURN_MEMORY_ERROR;
+        }
+        *upstream_list = upstreams;
+    }
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_edns_maximum_udp_payload_size(getdns_context *context,
+    uint16_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->edns_maximum_udp_payload_size;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_edns_extended_rcode(getdns_context *context,
+    uint8_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->edns_extended_rcode;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_edns_version(getdns_context *context, uint8_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->edns_version;
+    return GETDNS_RETURN_GOOD;
+}
+
+getdns_return_t
+getdns_context_get_edns_do_bit(getdns_context *context, uint8_t* value) {
+    RETURN_IF_NULL(context, GETDNS_RETURN_INVALID_PARAMETER);
+    RETURN_IF_NULL(value, GETDNS_RETURN_INVALID_PARAMETER);
+    *value = context->edns_do_bit;
+    return GETDNS_RETURN_GOOD;
+}
+
 /* context.c */
