@@ -49,6 +49,7 @@ struct ub_ctx;
 
 #define GETDNS_FN_RESOLVCONF "/etc/resolv.conf"
 #define GETDNS_FN_HOSTS      "/etc/hosts"
+#define GETDNS_TLS_PORT 1021
 
 enum filechgs { GETDNS_FCHG_ERRORS = -1
  , GETDNS_FCHG_NOERROR   = 0
@@ -60,6 +61,9 @@ enum filechgs { GETDNS_FCHG_ERRORS = -1
 typedef void (*getdns_update_callback) (struct getdns_context *,
     getdns_context_code_t);
 
+typedef void (*getdns_update_callback2) (struct getdns_context *,
+    getdns_context_code_t, void *userarg);
+
 /* internal use only for detecting changes to system files */
 struct filechg {
 	char *fn;
@@ -68,6 +72,14 @@ struct filechg {
 	struct stat *prevstat;
 };
 
+typedef enum getdns_base_transport {
+	GETDNS_TRANSPORT_NONE,
+	GETDNS_TRANSPORT_UDP,
+	GETDNS_TRANSPORT_TCP_SINGLE,
+	GETDNS_TRANSPORT_TCP,
+	GETDNS_TRANSPORT_TLS
+} getdns_base_transport_t;
+
 typedef struct getdns_upstream {
 	struct getdns_upstreams *upstreams;
 
@@ -80,6 +92,7 @@ typedef struct getdns_upstream {
 
 	/* For sharing a TCP socket to this upstream */
 	int                      fd;
+	SSL*                     tls_obj;
 	getdns_eventloop_event   event;
 	getdns_eventloop        *loop;
 	getdns_tcp_state         tcp;
@@ -88,6 +101,18 @@ typedef struct getdns_upstream {
 	getdns_network_req      *write_queue;
 	getdns_network_req      *write_queue_last;
 	getdns_rbtree_t          netreq_by_query_id;
+
+	/* EDNS cookies */
+	uint32_t secret;
+	uint8_t  client_cookie[8];
+	uint8_t  prev_client_cookie[8];
+	uint8_t  server_cookie[32];
+
+	unsigned has_client_cookie : 1;
+	unsigned has_prev_client_cookie : 1;
+	unsigned has_server_cookie : 1;
+	unsigned server_cookie_len : 5;
+
 } getdns_upstream;
 
 typedef struct getdns_upstreams {
@@ -118,8 +143,11 @@ struct getdns_context {
 	uint8_t edns_version;
 	uint8_t edns_do_bit;
 	int edns_maximum_udp_payload_size; /* -1 is unset */
+	SSL_CTX* tls_ctx;
 
-	getdns_update_callback update_callback;
+	getdns_update_callback  update_callback;
+	getdns_update_callback2 update_callback2;
+	void                   *update_userarg;
 
 	int processing;
 	int destroying;
@@ -203,4 +231,8 @@ int filechg_check(struct getdns_context *context, struct filechg *fchg);
 
 void priv_getdns_context_ub_read_cb(void *userarg);
 
+getdns_base_transport_t priv_get_base_transport(getdns_transport_t transport, int level);
+
+void priv_getdns_upstreams_dereference(getdns_upstreams *upstreams);
+
 #endif /* _GETDNS_CONTEXT_H_ */
