@@ -34,8 +34,6 @@
 #include "config.h"
 #include <fcntl.h>
 #include "stub.h"
-#include "gldns/rrdef.h"
-#include "gldns/str2wire.h"
 #include "gldns/gbuffer.h"
 #include "gldns/pkthdr.h"
 #include "context.h"
@@ -43,218 +41,178 @@
 #include "util-internal.h"
 #include "general.h"
 
-static int
-getdns_make_query_pkt_buf(const getdns_network_req *netreq, uint8_t *buf,
-    size_t *olen, uint16_t *omax_udp_payload_size)
+static time_t secret_rollover_time = 0;
+static uint32_t secret = 0;
+static uint32_t prev_secret = 0;
+
+static void
+rollover_secret()
 {
-	size_t len;
+	time_t now = 0;
 
-	getdns_dns_req *dnsreq  = netreq->owner;
-	getdns_context *context = dnsreq->context;
-	getdns_dict *extensions = dnsreq->extensions;
+	/* Create and roll server secrets */
+	if (time(&now) <= secret_rollover_time)
+		return;
 
-	int dnssec_return_status
-	    = is_extension_set(extensions, "dnssec_return_status");
-	int dnssec_return_only_secure
-	    = is_extension_set(extensions, "dnssec_return_only_secure");
-	int dnssec_return_validation_chain
-	    = is_extension_set(extensions, "dnssec_return_validation_chain");
-	int dnssec_extension_set = dnssec_return_status
-	    || dnssec_return_only_secure || dnssec_return_validation_chain;
+	/* Remember previous secret, in to keep answering on rollover
+	 * boundry with old cookie.
+	 */
+	prev_secret = secret;
+	secret = arc4random();
 
-	uint32_t edns_do_bit;
-	int      edns_maximum_udp_payload_size;
-	uint32_t get_edns_maximum_udp_payload_size;
-	uint32_t edns_extended_rcode;
-	uint32_t edns_version;
-
-	getdns_dict *add_opt_parameters;
-	int     have_add_opt_parameters;
-
-	getdns_list *options;
-	size_t      noptions = 0;
-	size_t       i;
-
-	getdns_dict    *option;
-	uint32_t        option_code;
-	getdns_bindata *option_data;
-	size_t opt_options_size = 0;
-
-	int with_opt;
-	int r;
-	size_t dname_len;
-	
-	have_add_opt_parameters = getdns_dict_get_dict(extensions,
-	    "add_opt_parameters", &add_opt_parameters) == GETDNS_RETURN_GOOD;
-
-	if (dnssec_extension_set) {
-		edns_maximum_udp_payload_size =
-		    netreq->upstream->addr.ss_family == AF_INET6 ? 1232 : 1432;
-		edns_extended_rcode = 0;
-		edns_version = 0;
-		edns_do_bit = 1;
-	} else {
-		edns_maximum_udp_payload_size =
-		    context->edns_maximum_udp_payload_size;
-		edns_extended_rcode = context->edns_extended_rcode;
-		edns_version = context->edns_version;
-		edns_do_bit = context->edns_do_bit;
-
-		if (have_add_opt_parameters) {
-			if (!getdns_dict_get_int(add_opt_parameters,
-			    "maximum_udp_payload_size",
-			    &get_edns_maximum_udp_payload_size))
-				edns_maximum_udp_payload_size =
-				    get_edns_maximum_udp_payload_size;
-			(void) getdns_dict_get_int(add_opt_parameters,
-			    "extended_rcode", &edns_extended_rcode);
-			(void) getdns_dict_get_int(add_opt_parameters,
-			    "version", &edns_version);
-			(void) getdns_dict_get_int(add_opt_parameters,
-			    "do_bit", &edns_do_bit);
-		}
-	}
-	if (have_add_opt_parameters && getdns_dict_get_list(
-	    add_opt_parameters, "options", &options) == GETDNS_RETURN_GOOD)
-		(void) getdns_list_get_length(options, &noptions);
-
-	with_opt = edns_do_bit != 0 || edns_maximum_udp_payload_size != -1 ||
-	    edns_extended_rcode != 0 || edns_version != 0 || noptions;
-
-	*omax_udp_payload_size = edns_maximum_udp_payload_size =
-	    ! with_opt ? 512
-	    : edns_maximum_udp_payload_size == -1 ?
-	      netreq->upstream->addr.ss_family==AF_INET6 ? 1232 : 1432
-	    : edns_maximum_udp_payload_size > 512 ?
-	      edns_maximum_udp_payload_size : 512;
-
-	assert(buf);
-	assert(olen);
-
-	len = *olen;
-	*olen = 0;
-
-	if (len < GLDNS_HEADER_SIZE)
-		return GLDNS_WIREPARSE_ERR_BUFFER_TOO_SMALL;
-
-	gldns_write_uint16(buf + 2, 0); /* reset all flags */
-	GLDNS_RD_SET(buf);
-	if (dnssec_extension_set) /* We will do validation outselves */
-		GLDNS_CD_SET(buf);
-	GLDNS_OPCODE_SET(buf, GLDNS_PACKET_QUERY);
-	gldns_write_uint16(buf + GLDNS_QDCOUNT_OFF, 1); /* 1 query */
-	gldns_write_uint16(buf + GLDNS_ANCOUNT_OFF, 0); /* 0 answers */
-	gldns_write_uint16(buf + GLDNS_NSCOUNT_OFF, 0); /* 0 authorities */
-	gldns_write_uint16(buf + GLDNS_ARCOUNT_OFF, with_opt ? 1 : 0);
-
-	len   -= GLDNS_HEADER_SIZE;
-	*olen += GLDNS_HEADER_SIZE;
-	buf   += GLDNS_HEADER_SIZE;
-
-	dname_len = len;
-	if ((r = gldns_str2wire_dname_buf(dnsreq->name, buf, &dname_len)))
-		return r;
-	len   -= dname_len;
-	*olen += dname_len;
-	buf   += dname_len;
-
-	if (len < 4)
-		return GLDNS_WIREPARSE_ERR_BUFFER_TOO_SMALL;
-	gldns_write_uint16(buf, netreq->request_type);
-	gldns_write_uint16(buf + 2, netreq->request_class);
-	len   -= 4;
-	*olen += 4;
-	buf   += 4;
-
-	if (with_opt) {
-		if (len < 11)
-			return GLDNS_WIREPARSE_ERR_BUFFER_TOO_SMALL;
-
-		*omax_udp_payload_size = edns_maximum_udp_payload_size;
-		buf[0] = 0; /* dname for . */
-		gldns_write_uint16(buf + 1, GLDNS_RR_TYPE_OPT);
-		gldns_write_uint16(buf + 3,
-		    (uint16_t) edns_maximum_udp_payload_size);
-		buf[5] = (uint8_t) edns_extended_rcode;
-		buf[6] = (uint8_t) edns_version;
-		buf[7] = edns_do_bit ? 0x80 : 0;
-		buf[8] = 0;
-		gldns_write_uint16(buf + 9, (uint16_t) opt_options_size);
-		len   -= 11;
-		*olen += 11;
-		buf   += 11;
-		for (i = 0; i < noptions; i++) {
-			if (getdns_list_get_dict(options, i, &option))
-			    continue;
-			if (getdns_dict_get_int(
-			    option, "option_code", &option_code)) continue;
-			if (getdns_dict_get_bindata(
-			    option, "option_data", &option_data)) continue;
-
-			if (len < option_data->size + 4) {
-				gldns_write_uint16(buf - opt_options_size - 2,
-				    (uint16_t) opt_options_size);
-				return GLDNS_WIREPARSE_ERR_BUFFER_TOO_SMALL;
-			}
-			gldns_write_uint16(buf, (uint16_t) option_code);
-			gldns_write_uint16(buf + 2,
-			    (uint16_t) option_data->size);
-			(void) memcpy(buf + 4, option_data->data,
-			    option_data->size);
-
-			opt_options_size += option_data->size + 4;
-			len              -= option_data->size + 4;
-			*olen            += option_data->size + 4;
-			buf              += option_data->size + 4;
-		}
-		gldns_write_uint16(buf - opt_options_size - 2,
-		    (uint16_t) opt_options_size);
-	}
-	return 0;
+	/* Next rollover over EDNS_COOKIE_ROLLOVER_TIME with 30% jitter,
+	 * I.e. some offset + or - 15% of the future point in time.
+	 */
+	secret_rollover_time = now + (EDNS_COOKIE_ROLLOVER_TIME / 20 * 17)
+		 + arc4random_uniform(EDNS_COOKIE_ROLLOVER_TIME / 10 * 3);
 }
 
-/* Return a rough estimate for mallocs */
-static size_t
-getdns_get_query_pkt_size(getdns_context *context,
-    const char *name, uint16_t request_type, getdns_dict *extensions)
+static void
+calc_new_cookie(getdns_upstream *upstream, uint8_t *cookie)
 {
-	getdns_dict *add_opt_parameters;
+        const EVP_MD *md;
+        EVP_MD_CTX *mdctx;
+        unsigned char md_value[EVP_MAX_MD_SIZE];
+        unsigned int md_len;
+        size_t i;
+        sa_family_t af = upstream->addr.ss_family;
+        void *sa_addr = ((struct sockaddr*)&upstream->addr)->sa_data;
+	size_t addr_len = ( af == AF_INET6 ? sizeof(struct sockaddr_in6)
+	                  : af == AF_INET  ? sizeof(struct sockaddr_in)
+	                  : 0 ) - sizeof(sa_family_t);
 
-	getdns_list *options;
-	size_t      noptions = 0;
-	size_t       i;
+        md = EVP_sha256();
+        mdctx = EVP_MD_CTX_create();
+        EVP_DigestInit_ex(mdctx, md, NULL);
+        EVP_DigestUpdate(mdctx, &secret, sizeof(secret));
+        EVP_DigestUpdate(mdctx, sa_addr, addr_len);
+        EVP_DigestFinal_ex(mdctx, md_value, &md_len);
+        EVP_MD_CTX_destroy(mdctx);
 
-	getdns_dict    *option;
-	uint32_t        option_code;
-	getdns_bindata *option_data;
-	size_t opt_options_size = 0;
+        (void) memset(cookie, 0, 8);
+        for (i = 0; i < md_len; i++)
+                cookie[i % 8] ^= md_value[i];
+}
 
-	do {
-		if (getdns_dict_get_dict(extensions,
-		    "add_opt_parameters", &add_opt_parameters)) break;
-		if (getdns_dict_get_list(
-		    add_opt_parameters, "options", &options)) break;
-		if (getdns_list_get_length(options, &noptions)) break;
+static uint8_t *
+attach_edns_cookie(getdns_upstream *upstream, uint8_t *opt)
+{
+	rollover_secret();
 
-		for (i = 0; i < noptions; i++) {
-			if (getdns_list_get_dict(options, i, &option)) continue;
-			if (getdns_dict_get_int(
-			    option, "option_code", &option_code)) continue;
-			if (getdns_dict_get_bindata(
-			    option, "option_data", &option_data)) continue;
+	if (!upstream->has_client_cookie) {
+		calc_new_cookie(upstream, upstream->client_cookie);
+		upstream->secret = secret;
+		upstream->has_client_cookie = 1;
 
-			opt_options_size += option_data->size
-			    + 2 /* option-code   */
-			    + 2 /* option-length */
-			    ;
-		}
-	} while (0);
-	
-	return GLDNS_HEADER_SIZE
-	    + strlen(name) + 1 + 4 /* dname always smaller then strlen(name) + 1 */
-	    + 12 + opt_options_size /* space needed for OPT (if needed) */
-	    /* TODO: TSIG */
-	    ;
+		gldns_write_uint16(opt +  9, 12); /* rdata len */
+		gldns_write_uint16(opt + 11, EDNS_COOKIE_OPCODE);
+		gldns_write_uint16(opt + 13,  8); /* opt len */
+		memcpy(opt + 15, upstream->client_cookie, 8);
+		return opt + 23;
+
+	} else if (upstream->secret != secret) {
+		memcpy( upstream->prev_client_cookie
+		      , upstream->client_cookie, 8);
+		upstream->has_prev_client_cookie = 1;
+		calc_new_cookie(upstream, upstream->client_cookie);
+		upstream->secret = secret;
+
+		gldns_write_uint16(opt +  9, 12); /* rdata len */
+		gldns_write_uint16(opt + 11, EDNS_COOKIE_OPCODE);
+		gldns_write_uint16(opt + 13,  8); /* opt len */
+		memcpy(opt + 15, upstream->client_cookie, 8);
+		return opt + 23;
+
+	} else if (!upstream->has_server_cookie) {
+		gldns_write_uint16(opt +  9, 12); /* rdata len */
+		gldns_write_uint16(opt + 11, EDNS_COOKIE_OPCODE);
+		gldns_write_uint16(opt + 13,  8); /* opt len */
+		memcpy(opt + 15, upstream->client_cookie, 8);
+		return opt + 23;
+	} else {
+		gldns_write_uint16( opt +  9, 12  /* rdata len */
+		                  + upstream->server_cookie_len);
+		gldns_write_uint16(opt + 11, EDNS_COOKIE_OPCODE);
+		gldns_write_uint16(opt + 13,  8   /* opt len */
+		                  + upstream->server_cookie_len);
+		memcpy(opt + 15, upstream->client_cookie, 8);
+		memcpy(opt + 23, upstream->server_cookie
+		               , upstream->server_cookie_len);
+		return opt + 23+ upstream->server_cookie_len;
+	}
+}
+
+static int
+match_and_process_server_cookie(
+    getdns_upstream *upstream, uint8_t *response, size_t response_len)
+{
+	priv_getdns_rr_iter rr_iter_storage, *rr_iter;
+	uint8_t *pos;
+	uint16_t rdata_len, opt_code, opt_len;
+
+	/* Search for the OPT RR (if any) */
+	for ( rr_iter = priv_getdns_rr_iter_init(&rr_iter_storage
+	                                        , response, response_len)
+	    ; rr_iter
+	    ; rr_iter = priv_getdns_rr_iter_next(rr_iter)) {
+
+		if (priv_getdns_rr_iter_section(rr_iter) !=
+		    GLDNS_SECTION_ADDITIONAL)
+			continue;
+
+		if (gldns_read_uint16(rr_iter->rr_type) != GETDNS_RRTYPE_OPT)
+			continue;
+
+		break;
+	}
+	if (! rr_iter)
+		return 0; /* No OPT, no cookie */
+
+	pos = rr_iter->rr_type + 8;
+
+	/* OPT found, now search for the cookie option */
+	if (pos + 2 > rr_iter->nxt)
+		return 1; /* FORMERR */
+
+	rdata_len = gldns_read_uint16(pos); pos += 2;
+	if (pos + rdata_len > rr_iter->nxt)
+		return 1; /* FORMERR */
+
+	while (pos < rr_iter->nxt) {
+		opt_code = gldns_read_uint16(pos); pos += 2;
+		opt_len  = gldns_read_uint16(pos); pos += 2;
+		if (pos + opt_len > rr_iter->nxt)
+			return 1; /* FORMERR */
+		if (opt_code == EDNS_COOKIE_OPCODE)
+			break;
+		pos += opt_len; /* Skip unknown options */
+	}
+	if (pos >= rr_iter->nxt || opt_code != EDNS_COOKIE_OPCODE)
+		return 0; /* Everything OK, just no cookie found. */
+
+	if (opt_len < 16 || opt_len > 40)
+		return 1; /* FORMERR */
+
+	if (!upstream->has_client_cookie)
+		return 1; /* Cookie reply, but we didn't sent one */
+
+	if (memcmp(upstream->client_cookie, pos, 8) != 0) {
+		if (!upstream->has_prev_client_cookie)
+			return 1; /* Cookie didn't match */
+		if (memcmp(upstream->prev_client_cookie, pos, 8) != 0)
+			return 1; /* Previous cookie didn't match either */
+
+		upstream->has_server_cookie = 0;
+		return 0; /* Don't store server cookie, because it
+		           * is for our previous client cookie
+			   */
+	}
+	pos += 8;
+	opt_len -= 8;
+	upstream->has_server_cookie = 1;
+	upstream->server_cookie_len = opt_len;
+	(void) memcpy(upstream->server_cookie, pos, opt_len);
+	return 0;
 }
 
 /** best effort to set nonblocking */
@@ -300,7 +258,6 @@ stub_cleanup(getdns_network_req *netreq)
 
 	GETDNS_CLEAR_EVENT(dnsreq->loop, &netreq->event);
 
-	GETDNS_NULL_FREE(dnsreq->context->mf, netreq->tcp.write_buf);
 	GETDNS_NULL_FREE(dnsreq->context->mf, netreq->tcp.read_buf);
 
 	/* Nothing globally scheduled? Then nothing queued */
@@ -361,6 +318,13 @@ upstream_erred(getdns_upstream *upstream)
 		netreq->state = NET_REQ_FINISHED;
 		priv_getdns_check_dns_req_complete(netreq->owner);
 	}
+	/* TODO[TLS]: When we get an error (which is probably a timeout) and are 
+	 * using to keep connections open should we leave the connection up here? */
+	if (upstream->tls_obj) {
+		SSL_shutdown(upstream->tls_obj);
+		SSL_free(upstream->tls_obj);
+		upstream->tls_obj = NULL;
+	}
 	close(upstream->fd);
 	upstream->fd = -1;
 }
@@ -401,33 +365,33 @@ stub_udp_read_cb(void *userarg)
 	getdns_dns_req *dnsreq = netreq->owner;
 	getdns_upstream *upstream = netreq->upstream;
 
-	static size_t pkt_buf_len = 4096;
-	size_t        pkt_len = pkt_buf_len;
-	uint8_t       pkt_buf[pkt_buf_len];
-	uint8_t      *pkt = pkt_buf;
-
 	ssize_t       read;
 
 	GETDNS_CLEAR_EVENT(dnsreq->loop, &netreq->event);
 
-	if (netreq->max_udp_payload_size > pkt_buf_len) {
-		pkt_len = netreq->max_udp_payload_size;
-		if (!(pkt = GETDNS_XMALLOC(
-		    dnsreq->context->mf, uint8_t, pkt_len)))
-			goto done;
-	}
-	read = recvfrom(netreq->fd, pkt, pkt_len, 0, NULL, NULL);
+	read = recvfrom(netreq->fd, netreq->response,
+	    netreq->max_udp_payload_size + 1, /* If read == max_udp_payload_size
+	                                       * then all is good.  If read ==
+	                                       * max_udp_payload_size + 1, then
+	                                       * we receive more then requested!
+	                                       * i.e. overflow
+	                                       */
+	    0, NULL, NULL);
 	if (read == -1 && (errno = EAGAIN || errno == EWOULDBLOCK))
-		goto exit;
+		return;
 
 	if (read < GLDNS_HEADER_SIZE)
-		goto exit; /* Not DNS */
+		return; /* Not DNS */
 	
-	if (GLDNS_ID_WIRE(pkt) != netreq->query_id)
-		goto exit; /* Cache poisoning attempt ;) */
+	if (GLDNS_ID_WIRE(netreq->response) != netreq->query_id)
+		return; /* Cache poisoning attempt ;) */
+
+	if (netreq->owner->edns_cookies && match_and_process_server_cookie(
+	    upstream, netreq->response, read))
+		return; /* Client cookie didn't match? */
 
 	close(netreq->fd);
-	if (GLDNS_TC_WIRE(pkt) &&
+	if (GLDNS_TC_WIRE(netreq->response) &&
 	    dnsreq->context->dns_transport ==
 	    GETDNS_TRANSPORT_UDP_FIRST_AND_FALL_BACK_TO_TCP) {
 
@@ -447,9 +411,9 @@ stub_udp_read_cb(void *userarg)
 		    getdns_eventloop_event_init(&netreq->event, netreq,
 		    NULL, stub_tcp_write_cb, stub_timeout_cb));
 
-		goto exit;
+		return;
 	}
-	ldns_wire2pkt(&(netreq->result), pkt, (size_t)read);
+	netreq->response_len = read;
 	dnsreq->upstreams->current = 0;
 
 	/* TODO: DNSSEC */
@@ -457,57 +421,43 @@ stub_udp_read_cb(void *userarg)
 	netreq->bogus  = 0;
 done:
 	netreq->state = NET_REQ_FINISHED;
-exit:
-	if (pkt && pkt != pkt_buf)
-		GETDNS_FREE(dnsreq->context->mf, pkt);
-	if (netreq->state == NET_REQ_FINISHED)
-		priv_getdns_check_dns_req_complete(dnsreq);
+	priv_getdns_check_dns_req_complete(dnsreq);
 }
 
 static void
 stub_udp_write_cb(void *userarg)
 {
 	getdns_network_req *netreq = (getdns_network_req *)userarg;
-	getdns_dns_req *dnsreq = netreq->owner;
-
-	static size_t   pkt_buf_len = 4096;
-	uint8_t         pkt_buf[pkt_buf_len];
-	uint8_t        *pkt = pkt_buf;
-	size_t          pkt_len;
-	size_t          pkt_size_needed;
+	getdns_dns_req     *dnsreq = netreq->owner;
+	size_t             pkt_len = netreq->response - netreq->query;
 
 	GETDNS_CLEAR_EVENT(dnsreq->loop, &netreq->event);
 
-	pkt_size_needed = getdns_get_query_pkt_size(dnsreq->context,
-	    dnsreq->name, netreq->request_type, dnsreq->extensions);
+	netreq->query_id = arc4random();
+	GLDNS_ID_SET(netreq->query, netreq->query_id);
+	if (netreq->opt) {
+		if (netreq->edns_maximum_udp_payload_size == -1)
+			gldns_write_uint16(netreq->opt + 3,
+			    ( netreq->max_udp_payload_size =
+			      netreq->upstream->addr.ss_family == AF_INET6
+			    ? 1232 : 1432));
+		if (netreq->owner->edns_cookies) {
+			netreq->response = attach_edns_cookie(
+			    netreq->upstream, netreq->opt);
+			pkt_len = netreq->response - netreq->query;
+		}
+	}
 
-	if (pkt_size_needed > pkt_buf_len) {
-		pkt = GETDNS_XMALLOC(
-		    dnsreq->context->mf, uint8_t, pkt_size_needed);
-		pkt_len = pkt_size_needed;
-	} else
-		pkt_len = pkt_buf_len;
-
-	if (getdns_make_query_pkt_buf(netreq, pkt_buf, &pkt_len,
-	    &netreq->max_udp_payload_size))
-		goto exit;
-
-	netreq->query_id = ldns_get_random();
-	GLDNS_ID_SET(pkt, netreq->query_id);
-
-	if ((ssize_t)pkt_len != sendto(netreq->fd, pkt, pkt_len, 0,
+	if ((ssize_t)pkt_len != sendto(netreq->fd, netreq->query, pkt_len, 0,
 	    (struct sockaddr *)&netreq->upstream->addr,
 	                        netreq->upstream->addr_len)) {
 		close(netreq->fd);
-		goto exit;
+		return;
 	}
 	GETDNS_SCHEDULE_EVENT(
 	    dnsreq->loop, netreq->fd, dnsreq->context->timeout,
 	    getdns_eventloop_event_init(&netreq->event, netreq,
 	    stub_udp_read_cb, NULL, stub_timeout_cb));
-exit:
-	if (pkt && pkt != pkt_buf)
-		GETDNS_FREE(dnsreq->context->mf, pkt);
 }
 
 static getdns_upstream *
@@ -578,7 +528,7 @@ stub_tcp_read(int fd, getdns_tcp_state *tcp, struct mem_funcs *mf)
 	tcp->to_read  -= read;
 	tcp->read_pos += read;
 	
-	if (tcp->to_read > 0)
+	if ((int)tcp->to_read > 0)
 		return STUB_TCP_AGAIN;
 
 	read = tcp->read_pos - tcp->read_buf;
@@ -596,7 +546,7 @@ stub_tcp_read(int fd, getdns_tcp_state *tcp, struct mem_funcs *mf)
 				buf_size *= 2;
 
 			if (!(buf = GETDNS_XREALLOC(*mf,
-			    tcp->read_buf, uint8_t, tcp->read_buf_len)))
+			    tcp->read_buf, uint8_t, buf_size)))
 				return STUB_TCP_ERROR;
 
 			tcp->read_buf = buf;
@@ -630,9 +580,16 @@ stub_tcp_read_cb(void *userarg)
 		GETDNS_CLEAR_EVENT(dnsreq->loop, &netreq->event);
 		if (q != netreq->query_id)
 			return;
+		if (netreq->owner->edns_cookies &&
+		    match_and_process_server_cookie(
+		    netreq->upstream, netreq->tcp.read_buf,
+		    netreq->tcp.read_pos - netreq->tcp.read_buf))
+			return; /* Client cookie didn't match? */
 		netreq->state = NET_REQ_FINISHED;
-		ldns_wire2pkt(&(netreq->result), netreq->tcp.read_buf,
-		    netreq->tcp.read_pos - netreq->tcp.read_buf);
+		netreq->response = netreq->tcp.read_buf;
+		netreq->response_len =
+		    netreq->tcp.read_pos - netreq->tcp.read_buf;
+		netreq->tcp.read_buf = NULL;
 		dnsreq->upstreams->current = 0;
 
 		/* TODO: DNSSEC */
@@ -645,6 +602,187 @@ stub_tcp_read_cb(void *userarg)
 	}
 }
 
+/** wait for a socket to become ready */
+static int
+sock_wait(int sockfd)
+{
+	int ret;
+	fd_set fds;
+	FD_ZERO(&fds);
+	FD_SET(FD_SET_T sockfd, &fds);
+	/*TODO[TLS]: Pick up this timeout from the context*/
+	struct timeval timeout = {5, 0 };
+	ret = select(sockfd+1, NULL, &fds, NULL, &timeout);
+	if(ret == 0)
+		/* timeout expired */
+		return 0;
+	else if(ret == -1)
+		/* error */
+		return 0;
+	return 1;
+}
+
+static int
+sock_connected(int sockfd) 
+{
+	/* wait(write) until connected or error */
+	while(1) {
+		int error = 0;
+		socklen_t len = (socklen_t)sizeof(error);
+
+		if(!sock_wait(sockfd)) {
+			close(sockfd);
+			return -1;
+		}
+
+		/* check if there is a pending error for nonblocking connect */
+		if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (void*)&error, &len) < 0) {
+			error = errno; /* on solaris errno is error */
+		}
+		if (error == EINPROGRESS || error == EWOULDBLOCK)
+			continue; /* try again */
+		else if (error != 0) {
+			close(sockfd);
+			return -1;
+		}
+		/* connected */
+		break;
+	}
+	return sockfd;
+}
+
+/* The connection testing and handshake should be handled by integrating this 
+ * with the event loop framework, but for now just implement a standalone
+ * handshake method.*/
+static SSL*
+do_tls_handshake(getdns_dns_req *dnsreq, getdns_upstream *upstream) 
+{
+	/*Lets make sure the connection is up before we try a handshake*/
+	if (errno == EINPROGRESS && sock_connected(upstream->fd) == -1) {
+		return NULL;
+	}
+
+	/* Create SSL instance */
+	if (dnsreq->context->tls_ctx == NULL)
+		return NULL;
+	SSL* ssl = SSL_new(dnsreq->context->tls_ctx);
+	if(!ssl) {
+		return NULL;
+	}
+	/* Connect the SSL object with a file descriptor */
+	if(!SSL_set_fd(ssl, upstream->fd)) {
+		SSL_free(ssl);
+		return NULL;
+	}
+	SSL_set_connect_state(ssl);
+	(void) SSL_set_mode(ssl, SSL_MODE_AUTO_RETRY);
+
+	int r;
+	int want;
+	fd_set fds;
+	FD_ZERO(&fds);
+	FD_SET(upstream->fd, &fds);
+	struct timeval timeout = {dnsreq->context->timeout/1000, 0 };
+	while ((r = SSL_do_handshake(ssl)) != 1)
+	{
+		want = SSL_get_error(ssl, r);
+		switch (want) {
+			case SSL_ERROR_WANT_READ:
+				if (select(upstream->fd + 1, &fds, NULL, NULL, &timeout) == 0) {
+					SSL_free(ssl);
+					return NULL;
+				}
+				break;
+			case SSL_ERROR_WANT_WRITE:
+				if (select(upstream->fd + 1, NULL, &fds, NULL, &timeout) == 0) {
+					SSL_free(ssl);
+					return NULL;
+				}
+				break;
+			default: 
+				SSL_free(ssl);
+				return NULL;
+	   }
+	}
+	return ssl;
+}
+
+static int
+stub_tls_read(SSL* tls_obj, getdns_tcp_state *tcp, struct mem_funcs *mf)
+{
+	ssize_t  read;
+	uint8_t *buf;
+	size_t   buf_size;
+
+	if (!tcp->read_buf) {
+		/* First time tls read, create a buffer for reading */
+		if (!(tcp->read_buf = GETDNS_XMALLOC(*mf, uint8_t, 4096)))
+			return STUB_TCP_ERROR;
+
+		tcp->read_buf_len = 4096;
+		tcp->read_pos = tcp->read_buf;
+		tcp->to_read = 2; /* Packet size */
+	}
+
+	ERR_clear_error();
+	read = SSL_read(tls_obj, tcp->read_pos, tcp->to_read);
+	if (read <= 0) {
+		/* TODO[TLS]: Handle SSL_ERROR_WANT_WRITE which means handshake
+		   renegotiation. Need to keep handshake state to do that.*/
+		int want = SSL_get_error(tls_obj, read);
+		if (want == SSL_ERROR_WANT_READ) {
+			return STUB_TCP_AGAIN; /* read more later */
+		} else 
+			return STUB_TCP_ERROR;
+	}
+	tcp->to_read  -= read;
+	tcp->read_pos += read;
+
+	if ((int)tcp->to_read > 0)
+		return STUB_TCP_AGAIN;
+
+	read = tcp->read_pos - tcp->read_buf;
+	if (read == 2) {
+		/* Read the packet size short */
+		tcp->to_read = gldns_read_uint16(tcp->read_buf);
+
+		if (tcp->to_read < GLDNS_HEADER_SIZE)
+			return STUB_TCP_ERROR;
+
+		/* Resize our buffer if needed */
+		if (tcp->to_read > tcp->read_buf_len) {
+			buf_size = tcp->read_buf_len;
+			while (tcp->to_read > buf_size)
+				buf_size *= 2;
+		
+			if (!(buf = GETDNS_XREALLOC(*mf,
+			    tcp->read_buf, uint8_t, buf_size)))
+				return STUB_TCP_ERROR;
+		
+			tcp->read_buf = buf;
+			tcp->read_buf_len = buf_size;
+		}
+
+		/* Ready to start reading the packet */
+		tcp->read_pos = tcp->read_buf;
+		read = SSL_read(tls_obj, tcp->read_pos, tcp->to_read);
+		if (read <= 0) {
+			/* TODO[TLS]: Handle SSL_ERROR_WANT_WRITE which means handshake
+			   renegotiation. Need to keep handshake state to do that.*/
+			int want = SSL_get_error(tls_obj, read);
+			if (want == SSL_ERROR_WANT_READ) {
+				return STUB_TCP_AGAIN; /* read more later */
+			} else 
+				return STUB_TCP_ERROR;
+		}
+		tcp->to_read  -= read;
+		tcp->read_pos += read;
+		if ((int)tcp->to_read > 0)
+			return STUB_TCP_AGAIN;
+	}
+	return GLDNS_ID_WIRE(tcp->read_buf);
+}
+
 static void netreq_upstream_read_cb(void *userarg);
 static void netreq_upstream_write_cb(void *userarg);
 static void
@@ -657,8 +795,14 @@ upstream_read_cb(void *userarg)
 	uint16_t query_id;
 	intptr_t query_id_intptr;
 
-	switch ((q = stub_tcp_read(upstream->fd, &upstream->tcp,
-	                          &upstream->upstreams->mf))) {
+	if (upstream->tls_obj)
+		q = stub_tls_read(upstream->tls_obj, &upstream->tcp,
+		              &upstream->upstreams->mf);
+	else
+		q = stub_tcp_read(upstream->fd, &upstream->tcp,
+		             &upstream->upstreams->mf);
+
+	switch (q) {
 	case STUB_TCP_AGAIN:
 		return;
 
@@ -667,17 +811,24 @@ upstream_read_cb(void *userarg)
 		return;
 
 	default:
+
 		/* Lookup netreq */
 		query_id = (uint16_t) q;
 		query_id_intptr = (intptr_t) query_id;
 		netreq = (getdns_network_req *)getdns_rbtree_delete(
 		    &upstream->netreq_by_query_id, (void *)query_id_intptr);
-		if (! netreq) /* maybe canceled */
-			break;
+		if (! netreq) /* maybe canceled */ {
+			/* reset read buffer */
+			upstream->tcp.read_pos = upstream->tcp.read_buf;
+			upstream->tcp.to_read = 2;
+			return;
+		}
 
 		netreq->state = NET_REQ_FINISHED;
-		ldns_wire2pkt(&(netreq->result), upstream->tcp.read_buf,
-		    upstream->tcp.read_pos - upstream->tcp.read_buf);
+		netreq->response = upstream->tcp.read_buf;
+		netreq->response_len =
+		    upstream->tcp.read_pos - upstream->tcp.read_buf;
+		upstream->tcp.read_buf = NULL;
 		upstream->upstreams->current = 0;
 
 		/* TODO: DNSSEC */
@@ -686,10 +837,6 @@ upstream_read_cb(void *userarg)
 
 		stub_cleanup(netreq);
 
-		/* reset read buffer */
-		upstream->tcp.read_pos = upstream->tcp.read_buf;
-		upstream->tcp.to_read = 2;
-
 		/* More to read/write for syncronous lookups? */
 		if (netreq->event.read_cb) {
 			dnsreq = netreq->owner;
@@ -736,47 +883,15 @@ stub_tcp_write(int fd, getdns_tcp_state *tcp, getdns_network_req *netreq)
 {
 	getdns_dns_req *dnsreq = netreq->owner;
 
-	static size_t   pkt_buf_len = 4096;
-	uint8_t         pkt_buf[pkt_buf_len];
-	uint8_t        *pkt = pkt_buf;
-	size_t          pkt_len;
-	size_t          query_pkt_size;
-
+	size_t          pkt_len = netreq->response - netreq->query;
 	ssize_t         written;
 	uint16_t        query_id;
 	intptr_t        query_id_intptr;
 
 	/* Do we have remaining data that we could not write before?  */
 	if (! tcp->write_buf) {
-		/* No, this is an initial write.
-		 * Create packet and try to send
+		/* No, this is an initial write. Try to send
 		 */
-		query_pkt_size = getdns_get_query_pkt_size(dnsreq->context,
-		    dnsreq->name, netreq->request_type, dnsreq->extensions);
-
-		if (query_pkt_size + 2 > pkt_buf_len) {
-			/* Not enough space in out stack buffer.
-			 * Allocate a buffer on the heap.
-			 */
-			if (!(pkt = GETDNS_XMALLOC(dnsreq->context->mf,
-			    uint8_t, query_pkt_size + 2)))
-				return STUB_TCP_ERROR;
-
-			tcp->write_buf = pkt;
-			tcp->write_buf_len = query_pkt_size + 2;
-			tcp->written = 0;
-
-			pkt_len = query_pkt_size;
-		} else
-			pkt_len = pkt_buf_len - 2;
-
-		/* Construct query packet */
-		if (getdns_make_query_pkt_buf(netreq, pkt + 2, &pkt_len,
-		    &netreq->max_udp_payload_size))
-			return STUB_TCP_ERROR;
-
-		/* Prepend length short */
-		gldns_write_uint16(pkt, pkt_len);
 
 		/* Not keeping connections open? Then the first random number
 		 * will do as the query id.
@@ -785,32 +900,42 @@ stub_tcp_write(int fd, getdns_tcp_state *tcp, getdns_network_req *netreq)
 		 * the write_queue) for that upstream.  Register this netreq 
 		 * by query_id in the process.
 		 */
-		if (dnsreq->context->dns_transport !=
-		    GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN)
-
-			query_id = ldns_get_random();
+		if ((dnsreq->context->dns_transport == GETDNS_TRANSPORT_TCP_ONLY) ||
+			(dnsreq->context->dns_transport == GETDNS_TRANSPORT_UDP_ONLY) ||
+			(dnsreq->context->dns_transport == GETDNS_TRANSPORT_UDP_FIRST_AND_FALL_BACK_TO_TCP))
+			query_id = arc4random();
 		else do {
-			query_id = ldns_get_random();
+			query_id = arc4random();
 			query_id_intptr = (intptr_t)query_id;
 			netreq->node.key = (void *)query_id_intptr;
 
 		} while (!getdns_rbtree_insert(
 		    &netreq->upstream->netreq_by_query_id, &netreq->node));
 
-		GLDNS_ID_SET(pkt + 2, query_id);
+		GLDNS_ID_SET(netreq->query, query_id);
+		if (netreq->opt) {
+			/* no limits on the max udp payload size with tcp */
+			gldns_write_uint16(netreq->opt + 3, 65535);
 
+			if (netreq->owner->edns_cookies) {
+				netreq->response = attach_edns_cookie(
+				    netreq->upstream, netreq->opt);
+				pkt_len = netreq->response - netreq->query;
+				gldns_write_uint16(netreq->query - 2, pkt_len);
+			}
+		}
 		/* We have an initialized packet buffer.
 		 * Lets see how much of it we can write
 		 */
 #ifdef USE_TCP_FASTOPEN
 		/* We use sendto() here which will do both a connect and send */
-		written = sendto(fd, pkt, pkt_len + 2, MSG_FASTOPEN,
-					(struct sockaddr *)&(netreq->upstream->addr),
-					netreq->upstream->addr_len);
+		written = sendto(fd, netreq->query - 2, pkt_len + 2,
+		    MSG_FASTOPEN, (struct sockaddr *)&(netreq->upstream->addr),
+		    netreq->upstream->addr_len);
 		/* If pipelining we will find that the connection is already up so 
 		   just fall back to a 'normal' write. */
 		if (written == -1 && errno == EISCONN) 
-			written = write(fd, pkt, pkt_len + 2);
+			written = write(fd, netreq->query - 2, pkt_len + 2);
 
 		if ((written == -1 && (errno == EAGAIN ||
 		                       errno == EWOULDBLOCK ||
@@ -820,26 +945,17 @@ stub_tcp_write(int fd, getdns_tcp_state *tcp, getdns_network_req *netreq)
 		                       errno == EINPROGRESS)) ||
 		     written  < pkt_len + 2) {
 #else
-		written = write(fd, pkt, pkt_len + 2);
+		written = write(fd, netreq->query - 2, pkt_len + 2);
 		if ((written == -1 && (errno == EAGAIN ||
 		                       errno == EWOULDBLOCK)) ||
 		     written  < pkt_len + 2) {
 #endif
 			/* We couldn't write the whole packet.
-			 * We have to return with STUB_TCP_AGAIN, but if
-			 * the packet was on the stack only, we have to copy
-			 * it to heap space fist, because the stack will be
-			 * gone after return.
+			 * We have to return with STUB_TCP_AGAIN.
+			 * Setup tcp to track the state.
 			 */
-			if (!tcp->write_buf) {
-				/* Copy stack packet buffer to heap  */
-				if (!(tcp->write_buf = GETDNS_XMALLOC(
-				    dnsreq->context->mf,uint8_t,pkt_len + 2)))
-				 	return STUB_TCP_ERROR;
-				(void) memcpy(tcp->write_buf, pkt, pkt_len + 2);
-				tcp->write_buf_len = pkt_len + 2;
-			}
-			/* Because written could be -1 (and errno EAGAIN) */
+			tcp->write_buf = netreq->query - 2;
+			tcp->write_buf_len = pkt_len + 2;
 			tcp->written = written >= 0 ? written : 0;
 
 			return STUB_TCP_AGAIN;
@@ -848,7 +964,7 @@ stub_tcp_write(int fd, getdns_tcp_state *tcp, getdns_network_req *netreq)
 			return STUB_TCP_ERROR;
 
 		/* We were able to write everything!  Start reading. */
-		GETDNS_NULL_FREE(dnsreq->context->mf, tcp->write_buf);
+		return (int) query_id;
 
 	} else {/* if (! tcp->write_buf) */
 
@@ -867,13 +983,12 @@ stub_tcp_write(int fd, getdns_tcp_state *tcp, getdns_network_req *netreq)
 			/* Still more to send */
 			return STUB_TCP_AGAIN;
 
+		query_id = (int)GLDNS_ID_WIRE(tcp->write_buf + 2);
 		/* Done. Start reading */
-		query_id = GLDNS_ID_WIRE(tcp->write_buf + 2);
+		tcp->write_buf = NULL;
+		return query_id;
 
 	} /* if (! tcp->write_buf) */
-
-	GETDNS_NULL_FREE(dnsreq->context->mf, tcp->write_buf);
-	return (int) query_id;
 }
 
 static void
@@ -902,6 +1017,54 @@ stub_tcp_write_cb(void *userarg)
 	}
 }
 
+static int
+stub_tls_write(SSL* tls_obj, getdns_tcp_state *tcp, getdns_network_req *netreq)
+{
+	size_t          pkt_len = netreq->response - netreq->query;
+	ssize_t         written;
+	uint16_t        query_id;
+	intptr_t        query_id_intptr;
+
+	/* Do we have remaining data that we could not write before?  */
+	if (! tcp->write_buf) {
+		/* No, this is an initial write. Try to send
+		 */
+
+		 /* Find a unique query_id not already written (or in
+		 * the write_queue) for that upstream.  Register this netreq 
+		 * by query_id in the process.
+		 */
+		do {
+			query_id = ldns_get_random();
+			query_id_intptr = (intptr_t)query_id;
+			netreq->node.key = (void *)query_id_intptr;
+
+		} while (!getdns_rbtree_insert(
+		    &netreq->upstream->netreq_by_query_id, &netreq->node));
+
+		GLDNS_ID_SET(netreq->query, query_id);
+		if (netreq->opt)
+			/* no limits on the max udp payload size with tcp */
+			gldns_write_uint16(netreq->opt + 3, 65535);
+
+		/* We have an initialized packet buffer.
+		 * Lets see how much of it we can write */
+		
+		// TODO[TLS]: Handle error cases, partial writes, renegotiation etc.
+		ERR_clear_error();
+		written = SSL_write(tls_obj, netreq->query - 2, pkt_len + 2);
+		if (written <= 0)
+			return STUB_TCP_ERROR;
+
+		/* We were able to write everything!  Start reading. */
+		return (int) query_id;
+
+	} 
+
+	return STUB_TCP_ERROR;
+}
+
+
 static void
 upstream_write_cb(void *userarg)
 {
@@ -910,7 +1073,12 @@ upstream_write_cb(void *userarg)
 	getdns_dns_req *dnsreq = netreq->owner;
 	int q;
 
-	switch ((q = stub_tcp_write(upstream->fd, &upstream->tcp, netreq))) {
+	if (upstream->tls_obj)
+		q = stub_tls_write(upstream->tls_obj, &upstream->tcp, netreq);
+	else
+		q = stub_tcp_write(upstream->fd, &upstream->tcp, netreq);
+
+	switch (q) {
 	case STUB_TCP_AGAIN:
 		return;
 
@@ -982,6 +1150,58 @@ upstream_schedule_netreq(getdns_upstream *upstream, getdns_network_req *netreq)
 	}
 }
 
+static in_port_t
+get_port(struct sockaddr_storage* addr)
+{
+	return ntohs(addr->ss_family == AF_INET
+	    ? ((struct sockaddr_in *)addr)->sin_port
+	    : ((struct sockaddr_in6*)addr)->sin6_port);
+}
+
+static void
+set_port(struct sockaddr_storage* addr, in_port_t port)
+{
+	addr->ss_family == AF_INET
+	    ? (((struct sockaddr_in *)addr)->sin_port = htons(port))
+	    : (((struct sockaddr_in6*)addr)->sin6_port = htons(port));
+}
+
+static int
+tcp_connect (getdns_upstream *upstream, getdns_base_transport_t transport) {
+
+	int fd =-1;
+	struct sockaddr_storage  connect_addr;
+	struct sockaddr_storage* addr = &upstream->addr;
+	socklen_t addr_len = upstream->addr_len;
+
+	/* TODO[TLS]: For now, override the port to a hardcoded value*/
+	if (transport == GETDNS_TRANSPORT_TLS &&
+	    (int)get_port(addr) != GETDNS_TLS_PORT) {
+		connect_addr = upstream->addr;
+		addr = &connect_addr;
+		set_port(addr, GETDNS_TLS_PORT);
+	}
+
+	if ((fd = socket(addr->ss_family, SOCK_STREAM, IPPROTO_TCP)) == -1)
+		return -1;
+
+	getdns_sock_nonblock(fd);
+#ifdef USE_TCP_FASTOPEN
+	/* Leave the connect to the later call to sendto() if using TCP*/
+	if (transport == GETDNS_TRANSPORT_TCP || 
+	    transport == GETDNS_TRANSPORT_TCP_SINGLE)
+		return fd;
+#endif
+	if (connect(fd, (struct sockaddr *)addr,
+	    addr_len) == -1) {
+		if (errno != EINPROGRESS) {
+			close(fd);
+			return -1;
+		}
+	}
+	return fd;
+}
+
 getdns_return_t
 priv_getdns_submit_stub_request(getdns_network_req *netreq)
 {
@@ -991,9 +1211,13 @@ priv_getdns_submit_stub_request(getdns_network_req *netreq)
 	if (!upstream)
 	    	return GETDNS_RETURN_GENERIC_ERROR;
 
-	switch(dnsreq->context->dns_transport) {
-	case GETDNS_TRANSPORT_UDP_ONLY:
-	case GETDNS_TRANSPORT_UDP_FIRST_AND_FALL_BACK_TO_TCP:
+	// Work out the primary and fallback transport options
+	getdns_base_transport_t transport    = priv_get_base_transport(
+	                                       dnsreq->context->dns_transport,0);
+	getdns_base_transport_t fb_transport = priv_get_base_transport(
+	                                       dnsreq->context->dns_transport,1);
+	switch(transport) {
+	case GETDNS_TRANSPORT_UDP:
 
 		if ((netreq->fd = socket(
 		    upstream->addr.ss_family, SOCK_DGRAM, IPPROTO_UDP)) == -1)
@@ -1009,23 +1233,10 @@ priv_getdns_submit_stub_request(getdns_network_req *netreq)
 
 		return GETDNS_RETURN_GOOD;
 
-	case GETDNS_TRANSPORT_TCP_ONLY:
+	case GETDNS_TRANSPORT_TCP_SINGLE:
 
-		if ((netreq->fd = socket(
-		    upstream->addr.ss_family, SOCK_STREAM, IPPROTO_TCP)) == -1)
+		if ((netreq->fd = tcp_connect(upstream, transport)) == -1)
 			return GETDNS_RETURN_GENERIC_ERROR;
-		
-		getdns_sock_nonblock(netreq->fd);
-#ifdef USE_TCP_FASTOPEN
-		/* Leave the connect to the later call to sendto() */
-#else
-		if (connect(netreq->fd, (struct sockaddr *)&upstream->addr,
-		    upstream->addr_len) == -1 && errno != EINPROGRESS) {
-
-			close(netreq->fd);
-			return GETDNS_RETURN_GENERIC_ERROR;
-		}
-#endif
 		netreq->upstream = upstream;
 
 		GETDNS_SCHEDULE_EVENT(
@@ -1035,34 +1246,51 @@ priv_getdns_submit_stub_request(getdns_network_req *netreq)
 
 		return GETDNS_RETURN_GOOD;
 	
-	case GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN:
+	case GETDNS_TRANSPORT_TCP:
+	case GETDNS_TRANSPORT_TLS:
 		
 		/* In coming comments, "global" means "context wide" */
 
 		/* Are we the first? (Is global socket initialized?) */
 		if (upstream->fd == -1) {
-			/* We are the first. Make global socket and connect. */
-			if ((upstream->fd = socket(upstream->addr.ss_family,
-			    SOCK_STREAM, IPPROTO_TCP)) == -1)
-				return GETDNS_RETURN_GENERIC_ERROR;
-			
-			getdns_sock_nonblock(upstream->fd);
-#ifdef USE_TCP_FASTOPEN
-		/* Leave the connect to the later call to sendto() */
-#else
-			if (connect(upstream->fd,
-			    (struct sockaddr *)&upstream->addr,
-			    upstream->addr_len) == -1 && errno != EINPROGRESS){
+			/* TODO[TLS]: We should remember on the context if we had to fallback
+			 * for this upstream so when re-connecting from a dropped TCP 
+			 * connection we don't retry TLS. */
+			int fallback = 0;
 
-				close(upstream->fd);
-				upstream->fd = -1;
-				return GETDNS_RETURN_GENERIC_ERROR;
+			/* We are the first. Make global socket and connect. */
+			if ((upstream->fd = tcp_connect(upstream, transport)) == -1) {
+				if (fb_transport == GETDNS_TRANSPORT_NONE)
+					return GETDNS_RETURN_GENERIC_ERROR;
+				if ((upstream->fd = tcp_connect(upstream, fb_transport)) == -1)
+					return GETDNS_RETURN_GENERIC_ERROR;
+				fallback = 1; 
+			}
+			
+			/* Now do a handshake for TLS. Note waiting for this to succeed or 
+			 * timeout blocks the scheduling of any messages for this upstream*/
+			if (transport == GETDNS_TRANSPORT_TLS && (fallback == 0)) {
+				upstream->tls_obj = do_tls_handshake(dnsreq, upstream);
+				if (!upstream->tls_obj) {
+					if (fb_transport == GETDNS_TRANSPORT_NONE)
+						return GETDNS_RETURN_GENERIC_ERROR;
+					close(upstream->fd);
+					if ((upstream->fd = tcp_connect(upstream, fb_transport)) == -1)
+						return GETDNS_RETURN_GENERIC_ERROR;
+				}
 			}
-#endif
 			/* Attach to the global event loop
 			 * so it can do it's own scheduling
 			 */
 			upstream->loop = dnsreq->context->extension;
+		} else {
+			/* Cater for the case of the user downgrading and existing TLS
+			   connection to TCP for some reason...*/
+			if (transport == GETDNS_TRANSPORT_TCP && upstream->tls_obj) {
+				SSL_shutdown(upstream->tls_obj);
+				SSL_free(upstream->tls_obj);
+				upstream->tls_obj = NULL;
+			}
 		}
 		netreq->upstream = upstream;
 
