@@ -29,9 +29,23 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <dirent.h>
 #include <getdns/getdns.h>
 #include <getdns/getdns_extra.h>
 
+static int quiet = 0;
+static int batch_mode = 0;
+static char *query_file = NULL;
+static int json = 0;
+static char *the_root = ".";
+static char *name;
+static getdns_context *context;
+static getdns_dict *extensions;
+static uint16_t request_type = GETDNS_RRTYPE_NS;
+static int timeout, edns0_size;
+static int async = 0, interactive = 0;
+static enum { GENERAL, ADDRESS, HOSTNAME, SERVICE } calltype = GENERAL;
+
 int get_rrtype(const char *t);
 
 getdns_dict *
@@ -90,41 +104,114 @@ print_usage(FILE *out, const char *progname)
 	fprintf(out, "\t-b <bufsize>\tSet edns0 max_udp_payload size\n");
 	fprintf(out, "\t-D\tSet edns0 do bit\n");
 	fprintf(out, "\t-d\tclear edns0 do bit\n");
+	fprintf(out, "\t-F <filename>\tread the queries from the specified file\n");
 	fprintf(out, "\t-G\tgeneral lookup\n");
 	fprintf(out, "\t-H\thostname lookup. (<name> must be an IP address; <type> is ignored)\n");
 	fprintf(out, "\t-h\tPrint this help\n");
 	fprintf(out, "\t-i\tPrint api information\n");
 	fprintf(out, "\t-I\tInteractive mode (> 1 queries on same context)\n");
+	fprintf(out, "\t-j\tOutput json response dict\n");
+	fprintf(out, "\t-J\tPretty print json response dict\n");
+	fprintf(out, "\t-p\tPretty print response dict\n");
 	fprintf(out, "\t-r\tSet recursing resolution type\n");
 	fprintf(out, "\t-s\tSet stub resolution type (default = recursing)\n");
 	fprintf(out, "\t-S\tservice lookup (<type> is ignored)\n");
 	fprintf(out, "\t-t <timeout>\tSet timeout in miliseconds\n");
 	fprintf(out, "\t-T\tSet transport to TCP only\n");
 	fprintf(out, "\t-O\tSet transport to TCP only keep connections open\n");
+	fprintf(out, "\t-L\tSet transport to TLS only keep connections open\n");
+	fprintf(out, "\t-E\tSet transport to TLS with TCP fallback only keep connections open\n");
 	fprintf(out, "\t-u\tSet transport to UDP with TCP fallback\n");
 	fprintf(out, "\t-U\tSet transport to UDP only\n");
+	fprintf(out, "\t-B\tBatch mode. Schedule all messages before processing responses.\n");
+	fprintf(out, "\t-q\tQuiet mode - don't print response\n");
 }
 
 void callback(getdns_context *context, getdns_callback_type_t callback_type,
     getdns_dict *response, void *userarg, getdns_transaction_t trans_id)
 {
-	getdns_dict **response_ptr = (getdns_dict **)userarg;
+	char *response_str;
 
-	if (response)
-		*response_ptr = response;
+	if (callback_type == GETDNS_CALLBACK_COMPLETE) {
+		/* This is a callback with data */;
+		if (!quiet && (response_str = json ?
+		    getdns_print_json_dict(response, json == 1)
+		  : getdns_pretty_print_dict(response))) {
+
+			fprintf(stdout, "ASYNC response:\n%s\n", response_str);
+			free(response_str);
+		}
+		fprintf(stderr,
+			"The callback with ID %llu  was successfull.\n",
+			(unsigned long long)trans_id);
+
+	} else if (callback_type == GETDNS_CALLBACK_CANCEL)
+		fprintf(stderr,
+			"The callback with ID %llu was cancelled. Exiting.\n",
+			(unsigned long long)trans_id);
+	else
+		fprintf(stderr,
+			"The callback got a callback_type of %d. Exiting.\n",
+			callback_type);
+
+	getdns_dict_destroy(response);
+	response = NULL;
 }
 
-static char *the_root = ".";
-static char *name;
-static getdns_context *context;
-static getdns_dict *extensions;
-static uint16_t request_type = GETDNS_RRTYPE_NS;
-static int timeout, edns0_size;
-static int async = 0, interactive = 0;
-static enum { GENERAL, ADDRESS, HOSTNAME, SERVICE } calltype = GENERAL;
-
 #define CONTINUE ((getdns_return_t)-2)
 
+static getdns_return_t set_cookie(getdns_dict *exts, char *cookie)
+{
+	uint8_t data[40];
+	size_t i;
+	getdns_return_t r = GETDNS_RETURN_GENERIC_ERROR;
+	getdns_bindata bindata;
+
+	getdns_dict *opt_parameters = getdns_dict_create();
+	getdns_list *options = getdns_list_create();
+	getdns_dict *option = getdns_dict_create();
+
+	if (*cookie == '=')
+		cookie++;
+
+	for (i = 0; i < 40 && *cookie; i++) {
+		if (*cookie >= '0' && *cookie <= '9')
+			data[i] = (uint8_t)(*cookie - '0') << 4;
+		else if (*cookie >= 'a' && *cookie <= 'f')
+			data[i] = (uint8_t)(*cookie - 'a' + 10) << 4;
+		else if (*cookie >= 'A' && *cookie <= 'F')
+			data[i] = (uint8_t)(*cookie - 'A' + 10) << 4;
+		else
+			goto done;
+		cookie++;
+		if (*cookie >= '0' && *cookie <= '9')
+			data[i] |= (uint8_t)(*cookie - '0');
+		else if (*cookie >= 'a' && *cookie <= 'f')
+			data[i] |= (uint8_t)(*cookie - 'a' + 10);
+		else if (*cookie >= 'A' && *cookie <= 'F')
+			data[i] |= (uint8_t)(*cookie - 'A' + 10);
+		else
+			goto done;
+		cookie++;;
+	}
+	bindata.data = data;
+	bindata.size = i;
+	if ((r = getdns_dict_set_int(option, "option_code", 65001)))
+		goto done;
+	if ((r = getdns_dict_set_bindata(option, "option_data", &bindata)))
+		goto done;
+	if ((r = getdns_list_set_dict(options, 0, option)))
+		goto done;
+	if ((r = getdns_dict_set_list(opt_parameters, "options", options)))
+		goto done;
+	r = getdns_dict_set_dict(exts, "add_opt_parameters", opt_parameters);
+done:
+	getdns_dict_destroy(option);
+	getdns_list_destroy(options);
+	getdns_dict_destroy(opt_parameters);
+	return r;
+}
+
 getdns_return_t parse_args(int argc, char **argv)
 {
 	getdns_return_t r = GETDNS_RETURN_GOOD;
@@ -141,7 +228,14 @@ getdns_return_t parse_args(int argc, char **argv)
 			continue;
 
 		} else if (arg[0] == '+') {
-			if ((r = getdns_dict_set_int(extensions, arg+1,
+			if (arg[1] == 's' && arg[2] == 'i' && arg[3] == 't' &&
+			   (arg[4] == '=' || arg[4] == '\0')) {
+				if ((r = set_cookie(extensions, arg+4))) {
+					fprintf(stderr, "Could not set cookie:"
+					    " %d", r);
+					break;
+				}
+			} else if ((r = getdns_dict_set_int(extensions, arg+1,
 			    GETDNS_EXTENSION_TRUE))) {
 				fprintf(stderr, "Could not set extension "
 				    "\"%s\": %d\n", argv[i], r);
@@ -196,6 +290,15 @@ getdns_return_t parse_args(int argc, char **argv)
 			case 'd':
 				(void) getdns_context_set_edns_do_bit(context, 0);
 				break;
+			case 'F':
+				if (c[1] != 0 || ++i >= argc || !*argv[i]) {
+					fprintf(stderr, "file name expected "
+					    "after -F\n");
+					return GETDNS_RETURN_GENERIC_ERROR;
+				}
+				query_file = argv[i];
+				interactive = 1;
+				break;
 			case 'G':
 				calltype = GENERAL;
 				break;
@@ -211,6 +314,17 @@ getdns_return_t parse_args(int argc, char **argv)
 			case 'I':
 				interactive = 1;
 				break;
+			case 'j':
+				json = 2;
+				break;
+			case 'J':
+				json = 1;
+				break;
+			case 'p':
+				json = 0;
+			case 'q':
+				quiet = 1;
+				break;
 			case 'r':
 				getdns_context_set_resolution_type(
 				    context,
@@ -247,6 +361,14 @@ getdns_return_t parse_args(int argc, char **argv)
 				getdns_context_set_dns_transport(context,
 				    GETDNS_TRANSPORT_TCP_ONLY_KEEP_CONNECTIONS_OPEN);
 				break;
+			case 'L':
+				getdns_context_set_dns_transport(context,
+				    GETDNS_TRANSPORT_TLS_ONLY_KEEP_CONNECTIONS_OPEN);
+				break;
+			case 'E':
+				getdns_context_set_dns_transport(context,
+				    GETDNS_TRANSPORT_TLS_FIRST_AND_FALL_BACK_TO_TCP_KEEP_CONNECTIONS_OPEN);
+				break;
 			case 'u':
 				getdns_context_set_dns_transport(context,
 				    GETDNS_TRANSPORT_UDP_FIRST_AND_FALL_BACK_TO_TCP);
@@ -255,6 +377,9 @@ getdns_return_t parse_args(int argc, char **argv)
 				getdns_context_set_dns_transport(context,
 				    GETDNS_TRANSPORT_UDP_ONLY);
 				break;
+			case 'B':
+				batch_mode = 1;
+			break;
 
 
 			default:
@@ -289,6 +414,7 @@ main(int argc, char **argv)
 	char *response_str;
 	getdns_return_t r;
 	getdns_dict *address = NULL;
+	FILE *fp = NULL;
 
 	name = the_root;
 	if ((r = getdns_context_create(&context, 1))) {
@@ -304,14 +430,28 @@ main(int argc, char **argv)
 	if ((r = parse_args(argc, argv)))
 		goto done_destroy_context;
 
+	if (query_file) {
+		fp = fopen(query_file, "rt");
+		if (fp == NULL) {
+			fprintf(stderr, "Could not open query file: %s\n", query_file);
+			goto done_destroy_context;
+		}
+	}
+
 	/* Make the call */
 	do {
 		char line[1024], *token, *linev[256];
 		int linec;
 		if (interactive) {
-			fprintf(stdout, "> ");
-			if (!fgets(line, 1024, stdin) || !*line)
-				break;
+			if (!query_file) {
+				fprintf(stdout, "> ");
+				if (!fgets(line, 1024, stdin) || !*line)
+					break;
+			} else {
+				if (!fgets(line, 1024, fp) || !*line)
+					break;
+				fprintf(stdout,"Found query: %s", line);
+			}
 
 			linev[0] = argv[0];
 			linec = 1;
@@ -358,8 +498,8 @@ main(int argc, char **argv)
 			}
 			if (r)
 				goto done_destroy_extensions;
-
-			getdns_context_run(context);
+			if (!batch_mode) 
+				getdns_context_run(context);
 		} else {
 			switch (calltype) {
 			case GENERAL:
@@ -384,16 +524,29 @@ main(int argc, char **argv)
 			}
 			if (r)
 				goto done_destroy_extensions;
-		}
-		if ((response_str = getdns_pretty_print_dict(response))) {
-			fprintf(stdout, "%s\n", response_str);
-			free(response_str);
-		} else {
-			r = GETDNS_RETURN_MEMORY_ERROR;
-			fprintf(stderr, "Could not print response\n");
+			if (!quiet) {
+				if ((response_str = json ?
+				    getdns_print_json_dict(response, json == 1)
+				  : getdns_pretty_print_dict(response))) {
+
+					fprintf( stdout, "SYNC response:\n%s\n"
+					       , response_str);
+					free(response_str);
+				} else {
+					r = GETDNS_RETURN_MEMORY_ERROR;
+					fprintf( stderr
+					       , "Could not print response\n");
+				}
+			} else if (r == GETDNS_RETURN_GOOD)
+				fprintf(stdout, "Response code was: GOOD\n");
+			else if (interactive)
+				fprintf(stderr, "An error occurred: %d\n", r);
 		}
 	} while (interactive);
 
+	if (batch_mode) 
+		getdns_context_run(context);
+
 	/* Clean up */
 done_destroy_extensions:
 	getdns_dict_destroy(extensions);
@@ -401,6 +554,9 @@ done_destroy_context:
 	if (response) getdns_dict_destroy(response);
 	getdns_context_destroy(context);
 
+	if (fp)
+		fclose(fp);
+
 	if (r == CONTINUE)
 		return 0;
 	if (r)
@@ -416,7 +572,8 @@ int get_rrtype(const char *t) {
 	case 'A':
 	case 'a': switch (t[1]) {
 	          case '\0': return GETDNS_RRTYPE_A;
-	          case '6': return GETDNS_RRTYPE_A6;
+	          case '6': if (t[2] == '\0') return GETDNS_RRTYPE_A6;
+                            return -1;
 	          case 'A':
 	          case 'a': /* before "AA", final "AA" (GETDNS_RRTYPE_AAAA) */
 	                    if ((t[2]|0x20) == 'a' && (t[3]|0x20) == 'a' && t[4] == '\0')
@@ -459,7 +616,8 @@ int get_rrtype(const char *t) {
 	                                        return GETDNS_RRTYPE_CDNSKEY;
 	                              return -1;
 	                    case 'S':
-	                    case 's': return GETDNS_RRTYPE_CDS;
+	                    case 's': if (t[3] == '\0') return GETDNS_RRTYPE_CDS;
+	                              return -1;
 	                    default : return -1;
 	                    };
 	          case 'E':
@@ -501,7 +659,8 @@ int get_rrtype(const char *t) {
 	                    default : return -1;
 	                    };
 	          case 'S':
-	          case 's': return GETDNS_RRTYPE_DS;
+	          case 's': if (t[2] == '\0') return GETDNS_RRTYPE_DS;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'E':
@@ -553,7 +712,8 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_HINFO;
 	                    return -1;
 	          case 'P':
-	          case 'p': return GETDNS_RRTYPE_HIP;
+	          case 'p': if (t[3] == '\0') return GETDNS_RRTYPE_HIP;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'I':
@@ -583,7 +743,8 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_KEY;
 	                    return -1;
 	          case 'X':
-	          case 'x': return GETDNS_RRTYPE_KX;
+	          case 'x': if (t[2] == '\0') return GETDNS_RRTYPE_KX;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'L':
@@ -602,7 +763,8 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_LOC;
 	                    return -1;
 	          case 'P':
-	          case 'p': return GETDNS_RRTYPE_LP;
+	          case 'p': if (t[2] == '\0') return GETDNS_RRTYPE_LP;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'M':
@@ -613,28 +775,36 @@ int get_rrtype(const char *t) {
 	                              return -1;
 	                    switch (t[4]) {
 	                    case 'A':
-	                    case 'a': return GETDNS_RRTYPE_MAILA;
+	                    case 'a': if (t[5] == '\0') return GETDNS_RRTYPE_MAILA;
+	                              return -1;
 	                    case 'B':
-	                    case 'b': return GETDNS_RRTYPE_MAILB;
+	                    case 'b': if (t[5] == '\0') return GETDNS_RRTYPE_MAILB;
+	                              return -1;
 	                    default : return -1;
 	                    };
 	          case 'B':
-	          case 'b': return GETDNS_RRTYPE_MB;
+	          case 'b': if (t[2] == '\0') return GETDNS_RRTYPE_MB;
+                            return -1;
 	          case 'D':
-	          case 'd': return GETDNS_RRTYPE_MD;
+	          case 'd': if (t[2] == '\0') return GETDNS_RRTYPE_MD;
+                            return -1;
 	          case 'F':
-	          case 'f': return GETDNS_RRTYPE_MF;
+	          case 'f': if (t[2] == '\0') return GETDNS_RRTYPE_MF;
+                            return -1;
 	          case 'G':
-	          case 'g': return GETDNS_RRTYPE_MG;
+	          case 'g': if (t[2] == '\0') return GETDNS_RRTYPE_MG;
+                            return -1;
 	          case 'I':
 	          case 'i': /* before "MI", final "NFO" (GETDNS_RRTYPE_MINFO) */
 	                    if ((t[2]|0x20) == 'n' && (t[3]|0x20) == 'f' && (t[4]|0x20) == 'o' && t[5] == '\0')
 	                              return GETDNS_RRTYPE_MINFO;
 	                    return -1;
 	          case 'R':
-	          case 'r': return GETDNS_RRTYPE_MR;
+	          case 'r': if (t[2] == '\0') return GETDNS_RRTYPE_MR;
+	                    return -1;
 	          case 'X':
-	          case 'x': return GETDNS_RRTYPE_MX;
+	          case 'x': if (t[2] == '\0') return GETDNS_RRTYPE_MX;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'N':
@@ -647,7 +817,8 @@ int get_rrtype(const char *t) {
 	          case 'I':
 	          case 'i': switch (t[2]) {
 	                    case 'D':
-	                    case 'd': return GETDNS_RRTYPE_NID;
+	                    case 'd': if (t[3] == '\0') return GETDNS_RRTYPE_NID;
+	                              return -1;
 	                    case 'M':
 	                    case 'm': /* before "NIM", final "LOC" (GETDNS_RRTYPE_NIMLOC) */
 	                              if ((t[3]|0x20) == 'l' && (t[4]|0x20) == 'o' && (t[5]|0x20) == 'c' && t[6] == '\0')
@@ -698,7 +869,8 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_OPENPGPKEY;
 	                    return -1;
 	          case 'T':
-	          case 't': return GETDNS_RRTYPE_OPT;
+	          case 't': if (t[3] == '\0') return GETDNS_RRTYPE_OPT;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'P':
@@ -709,7 +881,8 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_PTR;
 	                    return -1;
 	          case 'X':
-	          case 'x': return GETDNS_RRTYPE_PX;
+	          case 'x': if (t[2] == '\0') return GETDNS_RRTYPE_PX;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'R':
@@ -720,14 +893,16 @@ int get_rrtype(const char *t) {
 	                              return GETDNS_RRTYPE_RKEY;
 	                    return -1;
 	          case 'P':
-	          case 'p': return GETDNS_RRTYPE_RP;
+	          case 'p': if (t[2] == '\0') return GETDNS_RRTYPE_RP;
+	                    return -1;
 	          case 'R':
 	          case 'r': /* before "RR", final "SIG" (GETDNS_RRTYPE_RRSIG) */
 	                    if ((t[2]|0x20) == 's' && (t[3]|0x20) == 'i' && (t[4]|0x20) == 'g' && t[5] == '\0')
 	                              return GETDNS_RRTYPE_RRSIG;
 	                    return -1;
 	          case 'T':
-	          case 't': return GETDNS_RRTYPE_RT;
+	          case 't': if (t[2] == '\0') return GETDNS_RRTYPE_RT;
+	                    return -1;
 	          default : return -1;
 	          };
 	case 'S':
@@ -735,7 +910,8 @@ int get_rrtype(const char *t) {
 	          case 'I':
 	          case 'i': switch (t[2]) {
 	                    case 'G':
-	                    case 'g': return GETDNS_RRTYPE_SIG;
+	                    case 'g': if (t[3] == '\0') return GETDNS_RRTYPE_SIG;
+	                              return -1;
 	                    case 'N':
 	                    case 'n': /* before "SIN", final "K" (GETDNS_RRTYPE_SINK) */
 	                              if ((t[3]|0x20) == 'k' && t[4] == '\0')
@@ -806,7 +982,8 @@ int get_rrtype(const char *t) {
 	          case 'I':
 	          case 'i': switch (t[2]) {
 	                    case 'D':
-	                    case 'd': return GETDNS_RRTYPE_UID;
+	                    case 'd': if (t[3] == '\0') return GETDNS_RRTYPE_UID;
+	                              return -1;
 	                    case 'N':
 	                    case 'n': /* before "UIN", final "FO" (GETDNS_RRTYPE_UINFO) */
 	                              if ((t[3]|0x20) == 'f' && (t[4]|0x20) == 'o' && t[5] == '\0')
