@@ -36,8 +36,6 @@
 #ifndef TYPES_INTERNAL_H_
 #define TYPES_INTERNAL_H_
 
-#include <netinet/in.h>
-#include <ldns/ldns.h>
 #include "getdns/getdns.h"
 #include "getdns/getdns_extra.h"
 #include "util/rbtree.h"
@@ -47,7 +45,6 @@ struct getdns_context;
 struct getdns_upstreams;
 struct getdns_upstream;
 
-
 /**
  * \defgroup strings String Constants
  * @{
@@ -109,6 +106,7 @@ struct getdns_upstream;
 struct getdns_dns_req;
 struct getdns_network_req;
 
+typedef void (*internal_cb_t)(struct getdns_dns_req *dns_req);
 
 #define MF_PLAIN ((void *)&plain_mem_funcs_user_arg)
 extern void *plain_mem_funcs_user_arg;
@@ -187,13 +185,9 @@ typedef struct getdns_network_req
 	uint16_t request_class;
 
 	/* result */
-	ldns_pkt *result;
 	int secure;
 	int bogus;
 
-	/* next request to issue after this one */
-	struct getdns_network_req *next;
-
 	/* For stub resolving */
 	struct getdns_upstream *upstream;
 	int                     fd;
@@ -201,11 +195,24 @@ typedef struct getdns_network_req
 	getdns_tcp_state        tcp;
 	uint16_t                query_id;
 
+	int                     edns_maximum_udp_payload_size;
 	uint16_t                max_udp_payload_size;
 
 	/* Network requests scheduled to write after me */
 	struct getdns_network_req *write_queue_tail;
 
+	/* When more space is needed for the wire_data response than is
+	 * available in wire_data[], it will be allocated seperately.
+	 * response will then not point to wire_data anymore.
+	 */
+	size_t   query_len;
+	uint8_t *query;
+	uint8_t *opt; /* offset of OPT RR in query */
+	size_t   response_len;
+	uint8_t *response;
+	size_t   wire_data_sz;
+	uint8_t  wire_data[];
+
 } getdns_network_req;
 
 /**
@@ -222,17 +229,20 @@ typedef struct getdns_dns_req {
 	/* canceled flag */
 	int canceled;
 
-	/* current network request */
-	struct getdns_network_req *current_req;
-
-	/* first request in list */
-	struct getdns_network_req *first_req;
-
 	/* context that owns the request */
 	struct getdns_context *context;
 
 	/* request extensions */
-	struct getdns_dict *extensions;
+	int dnssec_return_status;
+	int dnssec_return_only_secure;
+	int dnssec_return_validation_chain;
+	int edns_cookies;
+
+	/* Internally used by return_validation_chain */
+	int dnssec_ok_checking_disabled;
+
+	/* internally scheduled request */
+	internal_cb_t internal_cb;
 
 	/* event loop */
 	getdns_eventloop *loop;
@@ -247,15 +257,22 @@ typedef struct getdns_dns_req {
 	/* for scheduling timeouts when using libunbound */
 	getdns_eventloop_event timeout;
 
-	/* dnssec status */
-	int return_dnssec_status;
-
 	/* mem funcs */
 	struct mem_funcs my_mf;
 
 	/* Stuff for stub resolving */
 	struct getdns_upstreams *upstreams;
 
+	/* network requests for this dns request.
+	 * The array is terminated with NULL.
+	 *
+	 * Memory for these netreqs has been allocated by the same malloc
+	 * operation that reserved space for this getdns_dns_req.
+	 * They will thus be freed as part of the desctruction of this struct,
+	 * and do not need to be freed seperately.
+	 */
+	getdns_network_req *netreqs[];
+
 } getdns_dns_req;
 
 #define GETDNS_XMALLOC(obj, type, count)	\
@@ -294,13 +311,6 @@ typedef struct getdns_dns_req {
 
 /* utility methods */
 
-/* network request utilities */
-void network_req_free(getdns_network_req * net_req);
-
-getdns_network_req *network_req_new(getdns_dns_req * owner,
-    uint16_t request_type,
-    uint16_t request_class, struct getdns_dict *extensions);
-
 /* dns request utils */
 getdns_dns_req *dns_req_new(getdns_context *context, getdns_eventloop *loop,
     const char *name, uint16_t request_type, getdns_dict *extensions);
